# generated by datamodel-codegen:
#   filename:  cardano.json
#   timestamp: 2024-08-24T16:37:37+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union
import sys

from pydantic.v1 import AnyUrl, Extra, Field, conint, constr

if "pytest" in sys.modules:
    from pydantic.v1 import BaseModel
else:
    from ogmios.model.BaseModelNoException import BaseModelNoException as BaseModel


class Cardano(BaseModel):
    __root__: Any = Field(..., title='Cardano')


class Address(BaseModel):
    __root__: constr(regex=r'[1-9A-HJ-NP-Za-km-z]*') = Field(
        ...,
        description='A Cardano address (either legacy format or new format).',
        examples=[
            'addr1q9d34spgg2kdy47n82e7x9pdd6vql6d2engxmpj20jmhuc2047yqd4xnh7u6u5jp4t0q3fkxzckph4tgnzvamlu7k5psuahzcp',
            'DdzFFzCqrht8mbSTZHqpM2u4HeND2mdspsaBhdQ1BowPJBMzbDeBMeKgqdoKqo1D4sdPusEdZJVrFJRBBxX1jUEofNDYCJSZLg8MkyCE',
        ],
        title='Address',
    )


class AddressAttributes(BaseModel):
    __root__: str = Field(
        ...,
        description='Extra attributes carried by Byron addresses (network magic and/or HD payload).',
        title='AddressAttributes',
    )


class DigestAny(BaseModel):
    __root__: str = Field(
        ...,
        description='A hash digest from an unspecified algorithm and length.',
        examples=['90181c517a5beadc9c3fe64bf8'],
        title='Digest<Any>',
    )


class AssetQuantity(BaseModel):
    __root__: int = Field(
        ...,
        description='A number of asset, can be negative went burning assets.',
        title='AssetQuantity',
    )


class Assets(BaseModel):
    __root__: Optional[Dict[str, Dict[str, AssetQuantity]]] = None


class Type(Enum):
    ebb = 'ebb'


class Era(Enum):
    byron = 'byron'


class DigestBlake2b256(BaseModel):
    __root__: constr(min_length=64, max_length=64) = Field(
        ...,
        description='A Blake2b 32-byte hash digest, encoded in base16.',
        examples=['c248757d390181c517a5beadc9c3fe64bf821d3e889a963fc717003ec248757d'],
        title='Digest<Blake2b, 256>',
    )


class Type1(Enum):
    bft = 'bft'


class BlockHeight(BaseModel):
    __root__: conint(ge=0, le=18446744073709552999) = Field(
        ...,
        description='A block number, the i-th block to be minted is number i.',
        title='BlockHeight',
    )


class Type2(Enum):
    praos = 'praos'


class Era2(Enum):
    shelley = 'shelley'
    allegra = 'allegra'
    mary = 'mary'
    alonzo = 'alonzo'
    babbage = 'babbage'


class Ancestor(Enum):
    genesis = 'genesis'


class VrfProof(BaseModel):
    __root__: str = Field(..., title='VrfProof')


class UInt32(BaseModel):
    __root__: conint(ge=0, le=4294967295) = Field(..., title='UInt32')


class VerificationKey(BaseModel):
    __root__: constr(min_length=64, max_length=64) = Field(
        ..., description='An Ed25519 verification key.', title='VerificationKey'
    )


class BootstrapProtocolId(BaseModel):
    __root__: conint(ge=0, le=4294967295) = Field(
        ..., examples=[764824073], title='BootstrapProtocolId'
    )


class Ratio(BaseModel):
    __root__: constr(regex=r'^-?[0-9]+/[0-9]+$') = Field(
        ...,
        description='A ratio of two integers, to express exact fractions.',
        examples=['2/3', '7/8'],
        title='Ratio',
    )


class Voter(BaseModel):
    class Config:
        extra = Extra.forbid

    verificationKey: Union[VerificationKey, ExtendedVerificationKey]  # Edited


class Proposal1(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b256


class BootstrapVote(BaseModel):
    class Config:
        extra = Extra.forbid

    voter: Voter
    proposal: Proposal1


class RelativeTime(BaseModel):
    class Config:
        extra = Extra.forbid

    seconds: int


class Type3(Enum):
    stakeDelegation = 'stakeDelegation'


class Type4(Enum):
    stakeCredentialRegistration = 'stakeCredentialRegistration'


class Type5(Enum):
    stakeCredentialDeregistration = 'stakeCredentialDeregistration'


class Type6(Enum):
    stakePoolRegistration = 'stakePoolRegistration'


class Type7(Enum):
    stakePoolRetirement = 'stakePoolRetirement'


class Type8(Enum):
    genesisDelegation = 'genesisDelegation'


class Type9(Enum):
    constitutionalCommitteeDelegation = 'constitutionalCommitteeDelegation'


class Type10(Enum):
    constitutionalCommitteeRetirement = 'constitutionalCommitteeRetirement'


class Type11(Enum):
    delegateRepresentativeRegistration = 'delegateRepresentativeRegistration'


class Type12(Enum):
    delegateRepresentativeUpdate = 'delegateRepresentativeUpdate'


class Type13(Enum):
    delegateRepresentativeRetirement = 'delegateRepresentativeRetirement'


class CredentialOrigin(Enum):
    verificationKey = 'verificationKey'
    script = 'script'


class ChainCode(BaseModel):
    __root__: str = Field(
        ..., description='An Ed25519-BIP32 chain-code for key deriviation.', title='ChainCode'
    )


class Null(BaseModel):
    __root__: None


class Status(Enum):
    active = 'active'
    expired = 'expired'
    unrecognized = 'unrecognized'


class Change(Enum):
    toBeEnacted = 'toBeEnacted'


class Next(BaseModel):
    class Config:
        extra = Extra.forbid

    change: Change


class Change1(Enum):
    toBeRemoved = 'toBeRemoved'


class Next1(BaseModel):
    class Config:
        extra = Extra.forbid

    change: Change1


class Change2(Enum):
    expiring = 'expiring'


class Next2(BaseModel):
    class Config:
        extra = Extra.forbid

    change: Change2


class Change3(Enum):
    adjustingMandate = 'adjustingMandate'


class DigestBlake2b224(BaseModel):
    __root__: constr(min_length=56, max_length=56) = Field(
        ...,
        description='A Blake2b 28-byte hash digest, encoded in base16.',
        examples=['90181c517a5beadc9c3fe64bf821d3e889a963fc717003ec248757d3'],
        title='Digest<Blake2b, 224>',
    )


class Status1(Enum):
    authorized = 'authorized'


class ConstitutionalCommitteeDelegate1(BaseModel):
    class Config:
        extra = Extra.forbid

    status: Status1
    id: DigestBlake2b224
    from_: CredentialOrigin = Field(..., alias='from')


class Status2(Enum):
    resigned = 'resigned'


class Status3(Enum):
    none = 'none'


class ConstitutionalCommitteeDelegate3(BaseModel):
    class Config:
        extra = Extra.forbid

    status: Status3


class Epoch(BaseModel):
    __root__: conint(ge=0, le=18446744073709552000) = Field(
        ..., description='An epoch number or length.', title='Epoch'
    )


class Int64(BaseModel):
    __root__: conint(ge=-9223372036854775808, le=9223372036854775807) = Field(..., title='Int64')


class Datum(BaseModel):
    __root__: str = Field(..., title='Datum')


class ConstitutionalCommittee(BaseModel):
    class Config:
        extra = Extra.forbid

    default: Ratio
    stateOfNoConfidence: Ratio


class ProtocolParametersUpdate(BaseModel):
    class Config:
        extra = Extra.forbid

    network: Ratio
    economic: Ratio
    technical: Ratio
    governance: Ratio


class DelegateRepresentativeVotingThresholds(BaseModel):
    class Config:
        extra = Extra.forbid

    noConfidence: Ratio
    constitution: Ratio
    constitutionalCommittee: ConstitutionalCommittee
    hardForkInitiation: Ratio
    protocolParametersUpdate: ProtocolParametersUpdate
    treasuryWithdrawals: Ratio


class Type14(Enum):
    registered = 'registered'


class DelegateRepresentative1(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b224
    from_: CredentialOrigin = Field(..., alias='from')
    type: Type14


class Type15(Enum):
    noConfidence = 'noConfidence'


class DelegateRepresentative2(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type15 = Field(
        ...,
        description='A special delegate representative which always vote no, except on votes of no-confidence.',
    )


class Type16(Enum):
    abstain = 'abstain'


class DelegateRepresentative3(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type16 = Field(..., description='A special delegate representative which always abstain.')


class DelegateRepresentative(BaseModel):
    __root__: Union[DelegateRepresentative1, DelegateRepresentative2, DelegateRepresentative3] = (
        Field(..., title='DelegateRepresentative')
    )


class Era3(Enum):
    byron = 'byron'
    shelley = 'shelley'
    allegra = 'allegra'
    mary = 'mary'
    alonzo = 'alonzo'
    babbage = 'babbage'
    conway = 'conway'


class EraWithGenesis(Enum):
    byron = 'byron'
    shelley = 'shelley'
    alonzo = 'alonzo'
    conway = 'conway'


class SlotLength(BaseModel):
    class Config:
        extra = Extra.forbid

    milliseconds: int


class UInt64(BaseModel):
    __root__: conint(ge=0, le=18446744073709552999) = Field(..., title='UInt64')


class Era4(Enum):
    byron = 'byron'


class UtcTime(BaseModel):
    __root__: datetime = Field(..., title='UtcTime')


class Era5(Enum):
    shelley = 'shelley'


class NetworkMagic(BaseModel):
    __root__: conint(ge=0, le=4294967296) = Field(
        ...,
        description='A magic number for telling networks apart. (e.g. 764824073)',
        examples=[764824073],
        title='NetworkMagic',
    )


class Era6(Enum):
    alonzo = 'alonzo'


class NumberOfBytes(BaseModel):
    class Config:
        extra = Extra.forbid

    bytes: UInt64


class Era7(Enum):
    conway = 'conway'


class ProtocolParametersUpdate1(BaseModel):
    class Config:
        extra = Extra.forbid

    security: Ratio


class StakePoolVotingThresholds(BaseModel):
    class Config:
        extra = Extra.forbid

    noConfidence: Ratio
    constitutionalCommittee: ConstitutionalCommittee
    hardForkInitiation: Ratio
    protocolParametersUpdate: ProtocolParametersUpdate1


class GenesisDelegate(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b224
    vrfVerificationKeyHash: DigestBlake2b256


class StakePoolId(BaseModel):
    __root__: constr(regex=r'^pool1[0-9a-z]*$') = Field(
        ...,
        description="A Blake2b 32-byte hash digest of a pool's verification key.",
        examples=[
            'pool1qqqqpanw9zc0rzh0yp247nzf2s35uvnsm7aaesfl2nnejaev0uc',
            'pool1qqqqqdk4zhsjuxxd8jyvwncf5eucfskz0xjjj64fdmlgj735lr9',
        ],
        title='StakePoolId',
    )


class Type17(Enum):
    protocolParametersUpdate = 'protocolParametersUpdate'


class Guardrails(BaseModel):
    class Config:
        extra = Extra.forbid

    hash: DigestBlake2b224


class Type18(Enum):
    hardForkInitiation = 'hardForkInitiation'


class Type19(Enum):
    treasuryTransfer = 'treasuryTransfer'


class Source(Enum):
    reserves = 'reserves'
    treasury = 'treasury'


class Target(Enum):
    reserves = 'reserves'
    treasury = 'treasury'


class Type20(Enum):
    treasuryWithdrawals = 'treasuryWithdrawals'


class Type21(Enum):
    constitutionalCommittee = 'constitutionalCommittee'


class RemovedItem(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b224
    from_: CredentialOrigin = Field(..., alias='from')


class Type22(Enum):
    constitution = 'constitution'


class Type23(Enum):
    noConfidence = 'noConfidence'


class Type24(Enum):
    information = 'information'


class GovernanceAction8(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type24


class TransactionId(BaseModel):
    __root__: constr(min_length=64, max_length=64) = Field(
        ...,
        description='A Blake2b 32-byte hash digest of a transaction body',
        title='TransactionId',
    )


class Ada(BaseModel):
    class Config:
        extra = Extra.forbid

    lovelace: int


class ValueAdaOnly(BaseModel):
    class Config:
        extra = Extra.forbid

    ada: Ada


class Vote(Enum):
    yes = 'yes'
    no = 'no'
    abstain = 'abstain'


class Role(Enum):
    genesisDelegate = 'genesisDelegate'


class GovernanceVoter1(BaseModel):
    class Config:
        extra = Extra.forbid

    role: Role
    id: DigestBlake2b224
    from_: CredentialOrigin = Field(..., alias='from')


class Role1(Enum):
    constitutionalCommittee = 'constitutionalCommittee'


class GovernanceVoter2(BaseModel):
    class Config:
        extra = Extra.forbid

    role: Role1
    id: DigestBlake2b224
    from_: CredentialOrigin = Field(..., alias='from')


class Role2(Enum):
    delegateRepresentative = 'delegateRepresentative'


class GovernanceVoter3(BaseModel):
    class Config:
        extra = Extra.forbid

    role: Role2
    id: DigestBlake2b224
    from_: CredentialOrigin = Field(..., alias='from')


class Role3(Enum):
    stakePoolOperator = 'stakePoolOperator'


class GovernanceVoter4(BaseModel):
    class Config:
        extra = Extra.forbid

    role: Role3
    id: StakePoolId


class GovernanceVoter(BaseModel):
    __root__: Union[GovernanceVoter1, GovernanceVoter2, GovernanceVoter3, GovernanceVoter4]


class ExtendedVerificationKey(BaseModel):
    __root__: constr(min_length=128, max_length=128) = Field(
        ...,
        description='An Ed25519-BIP32 Byron genesis delegate verification key with chain-code.',
        title='ExtendedVerificationKey',
    )


class Issuer4(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b224


class InitialDelegate(BaseModel):
    class Config:
        extra = Extra.forbid

    issuer: Issuer4
    delegate: GenesisDelegate


class InitialDelegates(BaseModel):
    __root__: List[InitialDelegate] = Field(..., title='InitialDelegates')


class InputSource(Enum):
    inputs = 'inputs'
    collaterals = 'collaterals'


class Language(Enum):
    plutus_v1 = 'plutus:v1'
    plutus_v2 = 'plutus:v2'
    plutus_v3 = 'plutus:v3'


class KesVerificationKey(BaseModel):
    __root__: str = Field(..., title='KesVerificationKey')


class Metadatum1(BaseModel):
    class Config:
        extra = Extra.forbid

    int: int


class Metadatum2(BaseModel):
    class Config:
        extra = Extra.forbid

    string: str


class Metadatum3(BaseModel):
    class Config:
        extra = Extra.forbid

    bytes: constr(regex=r'^[0-9a-f]*$')


class Field0(BaseModel):
    __root__: Union[int, str, List[Field0], Dict[str, Field0]] = Field(
        ..., title='Metadatum<NoSchema>'
    )


class Field11(BaseModel):
    class Config:
        extra = Extra.forbid

    int: int


class Field12(BaseModel):
    class Config:
        extra = Extra.forbid

    string: str


class Field13(BaseModel):
    class Config:
        extra = Extra.forbid

    bytes: constr(regex=r'^[0-9a-f]*$')


class Network(Enum):
    mainnet = 'mainnet'
    testnet = 'testnet'


class Nonce1(Enum):
    neutral = 'neutral'


class Nonce(BaseModel):
    __root__: Union[Nonce1, DigestBlake2b256] = Field(..., title='Nonce')


class Kes(BaseModel):
    class Config:
        extra = Extra.forbid

    period: UInt64
    verificationKey: KesVerificationKey


class Signature(BaseModel):
    __root__: constr(min_length=128, max_length=128) = Field(
        ..., description='An EdDSA signature.', title='Signature'
    )


class Origin(Enum):
    origin = 'origin'


class PolicyId(BaseModel):
    __root__: DigestBlake2b224 = Field(..., title='PolicyId')


class MinFeeReferenceScripts(BaseModel):
    class Config:
        extra = Extra.forbid

    range: UInt32
    base: float
    multiplier: float


class ScriptExecutionPrices(BaseModel):
    class Config:
        extra = Extra.forbid

    memory: Ratio
    cpu: Ratio


class MinFeeReferenceScripts1(BaseModel):
    class Config:
        extra = Extra.forbid

    range: UInt32
    base: float = Field(..., description='The base cost for the tier-price calculation.')
    multiplier: float = Field(..., description='The multiplier increasing the cost of each tier.')


class UInt16(BaseModel):
    __root__: conint(ge=0, le=65536) = Field(..., title='UInt16')


class RedeemerData(BaseModel):
    __root__: str = Field(..., description='Plutus data, CBOR-serialised.', title='RedeemerData')


class Purpose(Enum):
    spend = 'spend'
    mint = 'mint'
    publish = 'publish'
    withdraw = 'withdraw'
    vote = 'vote'
    propose = 'propose'


class RedeemerPointer(BaseModel):
    class Config:
        extra = Extra.forbid

    purpose: Purpose
    index: UInt64


class Type25(Enum):
    ipAddress = 'ipAddress'


class Relay1(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type25
    ipv4: Optional[str] = None
    ipv6: Optional[str] = None
    port: Optional[conint(ge=0, le=65535)] = None


class Type26(Enum):
    hostname = 'hostname'


class Relay2(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type26
    hostname: str
    port: Optional[conint(ge=0, le=65535)] = None


class Relay(BaseModel):
    __root__: Union[Relay1, Relay2] = Field(..., title='Relay')


class RewardAccount(BaseModel):
    __root__: constr(regex=r'^stake(_test)?1[0-9a-z]+$') = Field(
        ...,
        description="A reward account, also known as 'stake address'.",
        examples=['stake1ux7pt9adw8z46tgqn2f8fvurrhk325gcm4mf75mkmmxpx6gae9mzv'],
        title='RewardAccount',
    )


class Ada1(BaseModel):
    class Config:
        extra = Extra.forbid

    lovelace: conint(ge=-9223372036854775808, le=9223372036854775807) = Field(
        ..., description='An amount, possibly negative, in Lovelace (1e6 Lovelace = 1 Ada).'
    )


class ValueDelta(BaseModel):
    class Config:
        extra = Extra.forbid

    ada: Ada1


class SafeZone(BaseModel):
    __root__: conint(ge=0, le=18446744073709552999) = Field(
        ...,
        description='Number of slots from the tip of the ledger in which it is guaranteed that no hard fork can take place. This should be (at least) the number of slots in which we are guaranteed to have k blocks.',
        title='SafeZone',
    )


class Language1(Enum):
    native = 'native'


class Language2(Enum):
    plutus_v1 = 'plutus:v1'
    plutus_v2 = 'plutus:v2'
    plutus_v3 = 'plutus:v3'


class Script2(BaseModel):
    class Config:
        extra = Extra.forbid

    language: Language2
    cbor: str


class Clause(Enum):
    signature = 'signature'


class ScriptNative1(BaseModel):
    class Config:
        extra = Extra.forbid

    clause: Clause
    from_: DigestBlake2b224 = Field(..., alias='from')


class Clause1(Enum):
    any = 'any'


class Clause2(Enum):
    all = 'all'


class Clause3(Enum):
    some = 'some'


class Clause4(Enum):
    before = 'before'


class Clause5(Enum):
    after = 'after'


class Slot(BaseModel):
    __root__: conint(ge=0, le=18446744073709552000) = Field(
        ..., description='An absolute slot number.', title='Slot'
    )


class Purpose1(Enum):
    spend = 'spend'


class Purpose2(Enum):
    mint = 'mint'


class ScriptPurpose2(BaseModel):
    class Config:
        extra = Extra.forbid

    purpose: Purpose2
    policy: PolicyId


class Purpose3(Enum):
    publish = 'publish'


class Purpose4(Enum):
    withdraw = 'withdraw'


class ScriptPurpose4(BaseModel):
    class Config:
        extra = Extra.forbid

    purpose: Purpose4
    rewardAccount: RewardAccount


class Purpose5(Enum):
    propose = 'propose'


class Purpose6(Enum):
    vote = 'vote'


class ScriptPurpose6(BaseModel):
    class Config:
        extra = Extra.forbid

    purpose: Purpose6
    issuer: GovernanceVoter


class Transaction2(BaseModel):
    class Config:
        extra = Extra.forbid

    id: TransactionId


class TransactionOutputReference(BaseModel):
    class Config:
        extra = Extra.forbid

    transaction: Transaction2
    index: UInt64


class Signatory(BaseModel):
    class Config:
        extra = Extra.forbid

    key: VerificationKey
    signature: Signature
    chainCode: Optional[ChainCode] = None
    addressAttributes: Optional[AddressAttributes] = None


class SoftwareVersion(BaseModel):
    class Config:
        extra = Extra.forbid

    appName: str
    number: UInt32


class StakeAddress(BaseModel):
    __root__: constr(regex=r'^(stake|stake_test)1[0-9a-z]*$') = Field(
        ...,
        description='A stake address (a.k.a reward account)',
        examples=['stake179kzq4qulejydh045yzxwk4ksx780khkl4gdve9kzwd9vjcek9u8h'],
        title='StakeAddress',
    )


class Tip(BaseModel):
    class Config:
        extra = Extra.forbid

    slot: Slot
    id: DigestBlake2b256
    height: BlockHeight


class Treasury(BaseModel):
    class Config:
        extra = Extra.forbid

    value: Optional[ValueAdaOnly] = None
    donation: Optional[ValueAdaOnly] = None


class Ada2(BaseModel):
    class Config:
        extra = Extra.forbid

    lovelace: int


class Value(BaseModel):
    ada: Ada2


class UInt8(BaseModel):
    __root__: conint(ge=0, le=255) = Field(..., title='UInt8')


class ValidityInterval(BaseModel):
    class Config:
        extra = Extra.forbid

    invalidBefore: Optional[Slot] = None
    invalidAfter: Optional[Slot] = None


class VotingPeriod(Enum):
    voteForThisEpoch = 'voteForThisEpoch'
    voteForNextEpoch = 'voteForNextEpoch'


class VrfOutput(BaseModel):
    __root__: str = Field(..., title='VrfOutput')


class Withdrawals(BaseModel):
    __root__: Optional[Dict[str, ValueAdaOnly]] = None


class Anchor(BaseModel):
    class Config:
        extra = Extra.forbid

    hash: DigestAny
    url: AnyUrl


class BlockEBB(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type
    era: Era
    id: DigestBlake2b256
    ancestor: DigestBlake2b256
    height: BlockHeight


class Issuer(BaseModel):
    class Config:
        extra = Extra.forbid

    verificationKey: ExtendedVerificationKey


class Delegate(BaseModel):
    class Config:
        extra = Extra.forbid

    verificationKey: ExtendedVerificationKey


class CertifiedVrf(BaseModel):
    class Config:
        extra = Extra.forbid

    proof: Optional[VrfProof] = None
    output: Optional[VrfOutput] = None


class ProtocolVersion(BaseModel):
    class Config:
        extra = Extra.forbid

    major: UInt32
    minor: UInt32
    patch: Optional[UInt32] = None


class Issuer2(BaseModel):
    class Config:
        extra = Extra.forbid

    verificationKey: VerificationKey


class Delegate1(BaseModel):
    class Config:
        extra = Extra.forbid

    verificationKey: VerificationKey


class BootstrapOperationalCertificate(BaseModel):
    class Config:
        extra = Extra.forbid

    issuer: Issuer2
    delegate: Delegate1


class BootstrapProtocolParameters(BaseModel):
    class Config:
        extra = Extra.forbid

    heavyDelegationThreshold: Optional[Ratio] = None
    maxBlockBodySize: Optional[NumberOfBytes] = None
    maxBlockHeaderSize: Optional[NumberOfBytes] = None
    maxUpdateProposalSize: Optional[NumberOfBytes] = None
    maxTransactionSize: Optional[NumberOfBytes] = None
    multiPartyComputationThreshold: Optional[Ratio] = None
    scriptVersion: Optional[UInt64] = None
    slotDuration: Optional[UInt64] = None
    unlockStakeEpoch: Optional[UInt64] = None
    updateProposalThreshold: Optional[Ratio] = None
    updateProposalTimeToLive: Optional[UInt64] = None
    updateVoteThreshold: Optional[Ratio] = None
    softForkInitThreshold: Optional[Ratio] = None
    softForkMinThreshold: Optional[Ratio] = None
    softForkDecrementThreshold: Optional[Ratio] = None
    minFeeCoefficient: Optional[UInt64] = None
    minFeeConstant: Optional[ValueAdaOnly] = None


class Bound(BaseModel):
    class Config:
        extra = Extra.forbid

    time: RelativeTime
    slot: Slot
    epoch: Epoch


class StakePool(BaseModel):
    class Config:
        extra = Extra.forbid

    id: StakePoolId


class Certificate1(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type3
    from_: CredentialOrigin = Field(..., alias='from')
    credential: DigestBlake2b224
    stakePool: Optional[StakePool] = None
    delegateRepresentative: Optional[DelegateRepresentative] = None


class Certificate2(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type4
    from_: CredentialOrigin = Field(..., alias='from')
    credential: DigestBlake2b224
    deposit: Optional[ValueAdaOnly] = None


class Certificate3(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type5
    from_: CredentialOrigin = Field(..., alias='from')
    credential: DigestBlake2b224
    deposit: Optional[ValueAdaOnly] = None


class StakePool1(BaseModel):
    class Config:
        extra = Extra.forbid

    retirementEpoch: Epoch
    id: StakePoolId


class Certificate5(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type7
    stakePool: StakePool1


class Issuer3(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b224


class Certificate6(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type8
    delegate: GenesisDelegate
    issuer: Issuer3


class Member(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b224
    from_: CredentialOrigin = Field(..., alias='from')


class Certificate8(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type10
    member: Member
    metadata: Optional[Anchor] = None


class Certificate9(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type11
    delegateRepresentative: DelegateRepresentative
    deposit: ValueAdaOnly
    metadata: Optional[Anchor] = None


class Certificate10(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type12
    delegateRepresentative: DelegateRepresentative
    metadata: Optional[Anchor]


class Certificate11(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type13
    delegateRepresentative: DelegateRepresentative
    deposit: ValueAdaOnly


class Constitution(BaseModel):
    class Config:
        extra = Extra.forbid

    guardrails: Union[Null, Guardrails]
    metadata: Anchor


class ConstitutionalCommitteeDelegate2(BaseModel):
    class Config:
        extra = Extra.forbid

    status: Status2
    metadata: Optional[Anchor] = None


class ConstitutionalCommitteeDelegate(BaseModel):
    __root__: Union[
        ConstitutionalCommitteeDelegate1,
        ConstitutionalCommitteeDelegate2,
        ConstitutionalCommitteeDelegate3,
    ] = Field(..., title='ConstitutionalCommitteeDelegate')


class Mandate(BaseModel):
    class Config:
        extra = Extra.forbid

    epoch: Epoch


class CostModel(BaseModel):
    __root__: List[Int64] = Field(..., title='CostModel')


class EraParameters(BaseModel):
    class Config:
        extra = Extra.forbid

    epochLength: Epoch
    slotLength: SlotLength
    safeZone: Optional[SafeZone]


class EraSummary(BaseModel):
    class Config:
        extra = Extra.forbid

    start: Bound
    end: Optional[Bound] = None
    parameters: EraParameters


class ExecutionUnits(BaseModel):
    class Config:
        extra = Extra.forbid

    memory: UInt64
    cpu: UInt64


class UpdatableParameters1(BaseModel):
    class Config:
        extra = Extra.forbid

    stakePoolVotingThresholds: StakePoolVotingThresholds
    constitutionalCommitteeMinSize: UInt64
    constitutionalCommitteeMaxTermLength: UInt64
    governanceActionLifetime: Epoch
    governanceActionDeposit: ValueAdaOnly
    delegateRepresentativeVotingThresholds: DelegateRepresentativeVotingThresholds
    delegateRepresentativeDeposit: ValueAdaOnly
    delegateRepresentativeMaxIdleTime: Epoch


class StakePoolModel(BaseModel):
    class Config:
        extra = Extra.forbid

    id: StakePoolId
    vrfVerificationKeyHash: DigestBlake2b256
    owners: List[DigestBlake2b224]
    cost: ValueAdaOnly
    margin: Ratio
    pledge: ValueAdaOnly
    rewardAccount: RewardAccount
    metadata: Optional[Anchor] = None
    relays: List[Relay]


class GovernanceAction3(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type19
    source: Source
    target: Target
    value: ValueAdaOnly


class GovernanceProposalReference(BaseModel):
    class Config:
        extra = Extra.forbid

    transaction: Transaction2
    index: UInt32


class GovernanceVote(BaseModel):
    class Config:
        extra = Extra.forbid

    issuer: GovernanceVoter
    metadata: Optional[Anchor] = None
    vote: Vote
    proposal: Optional[GovernanceProposalReference] = None


class OperationalCertificate(BaseModel):
    class Config:
        extra = Extra.forbid

    count: UInt64
    sigma: Optional[Signature] = None
    kes: Kes


class Redeemer(BaseModel):
    class Config:
        extra = Extra.forbid

    redeemer: RedeemerData
    executionUnits: ExecutionUnits
    validator: RedeemerPointer


class RewardTransfer(BaseModel):
    __root__: Optional[Dict[str, ValueDelta]] = None


class ScriptNative5(BaseModel):
    class Config:
        extra = Extra.forbid

    clause: Clause4
    slot: Slot


class ScriptNative6(BaseModel):
    class Config:
        extra = Extra.forbid

    clause: Clause5
    slot: Slot


class ScriptPurpose1(BaseModel):
    class Config:
        extra = Extra.forbid

    purpose: Purpose1
    outputReference: TransactionOutputReference


class Protocol1(BaseModel):
    class Config:
        extra = Extra.forbid

    version: ProtocolVersion


class Issuer1(BaseModel):
    class Config:
        extra = Extra.forbid

    verificationKey: VerificationKey
    vrfVerificationKey: VerificationKey
    operationalCertificate: OperationalCertificate
    leaderValue: CertifiedVrf


class Proposal(BaseModel):
    class Config:
        extra = Extra.forbid

    version: ProtocolVersion
    software: SoftwareVersion
    parameters: BootstrapProtocolParameters
    metadata: Dict[str, str]


class BootstrapProtocolUpdate(BaseModel):
    class Config:
        extra = Extra.forbid

    proposal: Optional[Proposal] = None
    votes: List[BootstrapVote]


class Certificate4(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type6
    stakePool: StakePoolModel


class Certificate7(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type9
    member: Member
    delegate: ConstitutionalCommitteeDelegate


class Certificate(BaseModel):
    __root__: Union[
        Certificate1,
        Certificate2,
        Certificate3,
        Certificate4,
        Certificate5,
        Certificate6,
        Certificate7,
        Certificate8,
        Certificate9,
        Certificate10,
        Certificate11,
    ] = Field(..., title='Certificate')


class Next3(BaseModel):
    class Config:
        extra = Extra.forbid

    change: Change3
    mandate: Mandate


class ConstitutionalCommitteeMember(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b224
    from_: CredentialOrigin = Field(..., alias='from')
    status: Status = Field(
        ...,
        description="A member status. 'active' indicates that this member vote will count during the ratification of the ongoing epoch. 'unrecognized' means that some hot credential currently points to a non-existing (or no longer existing) member.",
    )
    delegate: ConstitutionalCommitteeDelegate
    mandate: Optional[Mandate] = None
    next: Optional[Union[Next, Next1, Next2, Next3]] = None


class ConstitutionalCommitteeMemberSummary(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b224
    from_: CredentialOrigin = Field(..., alias='from')
    mandate: Optional[Mandate] = None


class CostModels(BaseModel):
    __root__: Optional[Dict[str, CostModel]] = None


class UpdatableParameters(BaseModel):
    class Config:
        extra = Extra.forbid

    minUtxoDepositCoefficient: UInt64
    collateralPercentage: UInt64
    plutusCostModels: CostModels
    maxCollateralInputs: UInt64
    maxExecutionUnitsPerBlock: ExecutionUnits
    maxExecutionUnitsPerTransaction: ExecutionUnits
    maxValueSize: NumberOfBytes
    scriptExecutionPrices: ScriptExecutionPrices


class GenesisAlonzo(BaseModel):
    class Config:
        extra = Extra.forbid

    era: Era6
    updatableParameters: UpdatableParameters


class ConstitutionalCommittee1(BaseModel):
    class Config:
        extra = Extra.forbid

    members: List[ConstitutionalCommitteeMemberSummary]
    quorum: Ratio


class GenesisConway(BaseModel):
    class Config:
        extra = Extra.forbid

    era: Era7
    constitution: Constitution
    constitutionalCommittee: ConstitutionalCommittee1
    updatableParameters: UpdatableParameters1


class GenesisStakePools(BaseModel):
    class Config:
        extra = Extra.forbid

    stakePools: Dict[str, StakePoolModel]
    delegators: Dict[str, StakePoolId]


class GovernanceAction2(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type18
    ancestor: Optional[GovernanceProposalReference] = None
    version: ProtocolVersion


class GovernanceAction4(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type20
    withdrawals: RewardTransfer
    guardrails: Union[Null, Guardrails]


class Members(BaseModel):
    class Config:
        extra = Extra.forbid

    added: List[ConstitutionalCommitteeMemberSummary]
    removed: List[RemovedItem]


class GovernanceAction5(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type21
    ancestor: Optional[GovernanceProposalReference] = None
    members: Members
    quorum: Ratio


class GovernanceAction6(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type22
    ancestor: Optional[GovernanceProposalReference] = None
    guardrails: Union[Null, Guardrails]
    metadata: Anchor


class GovernanceAction7(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type23
    ancestor: Optional[GovernanceProposalReference] = None


class ProposedProtocolParameters(BaseModel):
    class Config:
        extra = Extra.forbid

    minFeeCoefficient: Optional[UInt64] = None
    minFeeConstant: Optional[ValueAdaOnly] = None
    minFeeReferenceScripts: Optional[MinFeeReferenceScripts] = None
    minUtxoDepositCoefficient: Optional[UInt64] = None
    minUtxoDepositConstant: Optional[ValueAdaOnly] = None
    maxBlockBodySize: Optional[NumberOfBytes] = None
    maxBlockHeaderSize: Optional[NumberOfBytes] = None
    maxTransactionSize: Optional[NumberOfBytes] = None
    maxReferenceScriptsSize: Optional[NumberOfBytes] = None
    maxValueSize: Optional[NumberOfBytes] = None
    extraEntropy: Optional[Nonce] = None
    stakeCredentialDeposit: Optional[ValueAdaOnly] = None
    stakePoolDeposit: Optional[ValueAdaOnly] = None
    stakePoolRetirementEpochBound: Optional[UInt64] = None
    stakePoolPledgeInfluence: Optional[Ratio] = None
    minStakePoolCost: Optional[ValueAdaOnly] = None
    desiredNumberOfStakePools: Optional[UInt64] = None
    federatedBlockProductionRatio: Optional[Ratio] = None
    monetaryExpansion: Optional[Ratio] = None
    treasuryExpansion: Optional[Ratio] = None
    collateralPercentage: Optional[UInt64] = None
    maxCollateralInputs: Optional[UInt64] = None
    plutusCostModels: Optional[CostModels] = None
    scriptExecutionPrices: Optional[ScriptExecutionPrices] = None
    maxExecutionUnitsPerTransaction: Optional[ExecutionUnits] = None
    maxExecutionUnitsPerBlock: Optional[ExecutionUnits] = None
    stakePoolVotingThresholds: Optional[StakePoolVotingThresholds] = None
    constitutionalCommitteeMinSize: Optional[UInt16] = None
    constitutionalCommitteeMaxTermLength: Optional[UInt64] = None
    governanceActionLifetime: Optional[Epoch] = None
    governanceActionDeposit: Optional[ValueAdaOnly] = None
    delegateRepresentativeVotingThresholds: Optional[DelegateRepresentativeVotingThresholds] = None
    delegateRepresentativeDeposit: Optional[ValueAdaOnly] = None
    delegateRepresentativeMaxIdleTime: Optional[Epoch] = None
    version: Optional[ProtocolVersion] = None


class ProtocolParameters(BaseModel):
    class Config:
        extra = Extra.forbid

    minFeeCoefficient: UInt64
    minFeeConstant: ValueAdaOnly
    minFeeReferenceScripts: Optional[MinFeeReferenceScripts1] = None
    minUtxoDepositCoefficient: UInt64
    minUtxoDepositConstant: ValueAdaOnly
    maxBlockBodySize: NumberOfBytes
    maxBlockHeaderSize: NumberOfBytes
    maxTransactionSize: Optional[NumberOfBytes] = None
    maxValueSize: Optional[NumberOfBytes] = None
    extraEntropy: Optional[Nonce] = None
    stakeCredentialDeposit: ValueAdaOnly
    stakePoolDeposit: ValueAdaOnly
    stakePoolRetirementEpochBound: UInt64
    stakePoolPledgeInfluence: Ratio
    minStakePoolCost: ValueAdaOnly
    desiredNumberOfStakePools: UInt64
    federatedBlockProductionRatio: Optional[Ratio] = None
    monetaryExpansion: Ratio
    treasuryExpansion: Ratio
    collateralPercentage: Optional[UInt64] = None
    maxCollateralInputs: Optional[UInt64] = None
    plutusCostModels: Optional[CostModels] = None
    scriptExecutionPrices: Optional[ScriptExecutionPrices] = None
    maxExecutionUnitsPerTransaction: Optional[ExecutionUnits] = None
    maxExecutionUnitsPerBlock: Optional[ExecutionUnits] = None
    maxReferenceScriptsSize: Optional[NumberOfBytes] = None
    stakePoolVotingThresholds: Optional[StakePoolVotingThresholds] = None
    constitutionalCommitteeMinSize: Optional[UInt16] = None
    constitutionalCommitteeMaxTermLength: Optional[UInt64] = None
    governanceActionLifetime: Optional[Epoch] = None
    governanceActionDeposit: Optional[ValueAdaOnly] = None
    delegateRepresentativeVotingThresholds: Optional[DelegateRepresentativeVotingThresholds] = None
    delegateRepresentativeDeposit: Optional[ValueAdaOnly] = None
    delegateRepresentativeMaxIdleTime: Optional[Epoch] = None
    version: ProtocolVersion


class ScriptPurpose3(BaseModel):
    class Config:
        extra = Extra.forbid

    purpose: Purpose3
    certificate: Certificate


class Protocol(BaseModel):
    class Config:
        extra = Extra.forbid

    id: BootstrapProtocolId
    version: ProtocolVersion
    software: SoftwareVersion
    update: Optional[BootstrapProtocolUpdate] = None


class GenesisByron(BaseModel):
    class Config:
        extra = Extra.forbid

    era: Era4
    genesisKeyHashes: List[DigestBlake2b224]
    genesisDelegations: Dict[str, BootstrapOperationalCertificate]
    startTime: UtcTime
    initialFunds: Dict[str, ValueAdaOnly]
    initialVouchers: Dict[str, ValueAdaOnly]
    securityParameter: UInt64
    networkMagic: NetworkMagic
    updatableParameters: Optional[ProtocolParameters] = None


class GenesisShelley(BaseModel):
    class Config:
        extra = Extra.forbid

    era: Era5
    startTime: UtcTime
    networkMagic: NetworkMagic
    network: Network
    activeSlotsCoefficient: Ratio
    securityParameter: UInt64
    epochLength: Epoch
    slotsPerKesPeriod: UInt64
    maxKesEvolutions: UInt64
    slotLength: SlotLength
    updateQuorum: UInt64
    maxLovelaceSupply: UInt64
    initialParameters: ProtocolParameters
    initialDelegates: InitialDelegates
    initialFunds: Dict[str, ValueAdaOnly] = Field(..., title='InitialFunds')
    initialStakePools: GenesisStakePools


class GovernanceAction1(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type17
    ancestor: Optional[GovernanceProposalReference] = None
    parameters: ProposedProtocolParameters
    guardrails: Union[Null, Guardrails]


class GovernanceAction(BaseModel):
    __root__: Union[
        GovernanceAction1,
        GovernanceAction2,
        GovernanceAction3,
        GovernanceAction4,
        GovernanceAction5,
        GovernanceAction6,
        GovernanceAction7,
        GovernanceAction8,
    ]


class GovernanceProposal(BaseModel):
    class Config:
        extra = Extra.forbid

    deposit: Optional[ValueAdaOnly] = None
    returnAccount: Optional[RewardAccount] = None
    metadata: Optional[Anchor] = None
    action: GovernanceAction


class ScriptPurpose5(BaseModel):
    class Config:
        extra = Extra.forbid

    purpose: Purpose5
    proposal: GovernanceProposal


class ScriptPurpose(BaseModel):
    __root__: Union[
        ScriptPurpose1,
        ScriptPurpose2,
        ScriptPurpose3,
        ScriptPurpose4,
        ScriptPurpose5,
        ScriptPurpose6,
    ] = Field(..., title='ScriptPurpose')


class Block(BaseModel):
    __root__: Union[BlockEBB, BlockBFT, BlockPraos] = Field(..., title='Block')


class BlockBFT(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type1
    era: Era
    id: DigestBlake2b256
    ancestor: DigestBlake2b256
    height: BlockHeight
    slot: Slot
    size: NumberOfBytes
    transactions: Optional[List[Transaction]] = None
    operationalCertificates: Optional[List[BootstrapOperationalCertificate]] = None
    protocol: Protocol
    issuer: Issuer
    delegate: Delegate


class BlockPraos(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type2
    era: Era2
    id: DigestBlake2b256
    ancestor: Union[DigestBlake2b256, Ancestor]
    nonce: Optional[CertifiedVrf] = None
    height: BlockHeight
    size: NumberOfBytes
    slot: Slot
    transactions: Optional[List[Transaction]] = None
    protocol: Protocol1
    issuer: Issuer1


class Metadata(BaseModel):
    class Config:
        extra = Extra.forbid

    hash: DigestBlake2b256
    labels: MetadataLabels


class MetadataLabels1(BaseModel):
    class Config:
        extra = Extra.forbid

    cbor: Optional[constr(regex=r'^[0-9a-f]*$')] = None
    json_: Optional[Metadatum] = Field(None, alias='json')


class MetadataLabels(BaseModel):
    __root__: Optional[Dict[str, MetadataLabels1]] = Field(
        None,
        description="An associated metadatum, as a CBOR bytestring or a JSON object if possible. If not possible, default to 'cbor'. Use --include-metadata-cbor to ALWAYS include the cbor field.",
    )


class Metadatum4(BaseModel):
    class Config:
        extra = Extra.forbid

    list: List[Field1]


class Metadatum5(BaseModel):
    class Config:
        extra = Extra.forbid

    map: List[MetadatumMap]


class Metadatum(BaseModel):
    __root__: Union[
        Union[int, str, List[Field0], Dict[str, Field0]],
        Union[Metadatum1, Metadatum2, Metadatum3, Metadatum4, Metadatum5],
    ] = Field(..., title='Metadatum')


class MetadatumMap(BaseModel):
    class Config:
        extra = Extra.forbid

    k: Field1
    v: Field1


class Field14(BaseModel):
    class Config:
        extra = Extra.forbid

    list: List[Field1]


class Field15(BaseModel):
    class Config:
        extra = Extra.forbid

    map: List[MetadatumMap]


class Field1(BaseModel):
    __root__: Union[Field11, Field12, Field13, Field14, Field15] = Field(
        ..., title='Metadatum<DetailedSchema>'
    )


class Script1(BaseModel):
    class Config:
        extra = Extra.forbid

    language: Language1
    json_: ScriptNative = Field(..., alias='json')
    cbor: Optional[str] = None


class Script(BaseModel):
    __root__: Union[Script1, Script2] = Field(..., title='Script')


class ScriptNative2(BaseModel):
    class Config:
        extra = Extra.forbid

    clause: Clause1
    from_: List[ScriptNative] = Field(..., alias='from')


class ScriptNative3(BaseModel):
    class Config:
        extra = Extra.forbid

    clause: Clause2
    from_: List[ScriptNative] = Field(..., alias='from')


class ScriptNative4(BaseModel):
    class Config:
        extra = Extra.forbid

    clause: Clause3
    atLeast: int
    from_: List[ScriptNative] = Field(..., alias='from')


class ScriptNative(BaseModel):
    __root__: Union[
        ScriptNative1, ScriptNative2, ScriptNative3, ScriptNative4, ScriptNative5, ScriptNative6
    ] = Field(
        ...,
        description='A phase-1 monetary script. Timelocks constraints are only supported since Allegra.',
        examples=[
            {
                'clause': 'signature',
                'from': '3c07030e36bfff7cd2f004356ef320f3fe3c07030e7cd2f004356437',
            },
            {
                'clause': 'all',
                'from': [
                    {
                        'clause': 'signature',
                        'from': 'ec09e5293d384637cd2f004356ef320f3fe3c07030e36bfffe67e2e2',
                    },
                    {
                        'clause': 'signature',
                        'from': '3c07030e36bfff7cd2f004356ef320f3fe3c07030e7cd2f004356437',
                    },
                ],
            },
            {
                'clause': 'some',
                'atLeast': 2,
                'from': [
                    {
                        'clause': 'signature',
                        'from': 'ec09e5293d384637cd2f004356ef320f3fe3c07030e36bfffe67e2e2',
                    },
                    {
                        'clause': 'signature',
                        'from': '3c07030e36bfff7cd2f004356ef320f3fe3c07030e7cd2f004356437',
                    },
                    {'clause': 'after', 'slot': 42},
                ],
            },
        ],
        title='Script<Native>',
    )


class Transaction(BaseModel):
    class Config:
        extra = Extra.forbid

    id: DigestBlake2b256
    spends: InputSource
    inputs: List[TransactionOutputReference]
    references: Optional[List[TransactionOutputReference]] = None
    collaterals: Optional[List[TransactionOutputReference]] = None
    totalCollateral: Optional[ValueAdaOnly] = None
    collateralReturn: Optional[TransactionOutput] = None
    outputs: List[TransactionOutput]
    certificates: Optional[List[Certificate]] = None
    withdrawals: Optional[Withdrawals] = None
    fee: Optional[ValueAdaOnly] = None
    validityInterval: Optional[ValidityInterval] = None
    mint: Optional[Assets] = None
    network: Optional[Network] = None
    scriptIntegrityHash: Optional[DigestBlake2b256] = None
    requiredExtraSignatories: Optional[List[DigestBlake2b224]] = None
    requiredExtraScripts: Optional[List[DigestBlake2b224]] = None
    proposals: Optional[List[GovernanceProposal]] = None
    votes: Optional[List[GovernanceVote]] = None
    metadata: Optional[Metadata] = None
    signatories: List[Signatory]
    scripts: Optional[Dict[str, Script]] = None
    datums: Optional[Dict[str, Datum]] = None
    redeemers: Optional[List[Redeemer]] = None
    treasury: Optional[Treasury] = None
    cbor: Optional[str] = Field(
        None,
        description="The raw serialized (CBOR) transaction, as found on-chain. Use --include-transaction-cbor to ALWAYS include the 'cbor' field. Omitted otherwise.",
    )


class TransactionOutput(BaseModel):
    class Config:
        extra = Extra.forbid

    address: Address
    value: Value
    datumHash: Optional[DigestBlake2b256] = None
    datum: Optional[Datum] = None
    script: Optional[Script] = None


class UtxoItem(BaseModel):
    class Config:
        extra = Extra.forbid

    transaction: Transaction2
    index: UInt32
    address: Address
    value: Value
    datumHash: Optional[DigestBlake2b256] = None
    datum: Optional[Datum] = None
    script: Optional[Script] = None


class Utxo(BaseModel):
    __root__: List[UtxoItem] = Field(..., title='Utxo')


Voter.update_forward_refs()  # Edited
Field0.update_forward_refs()
Block.update_forward_refs()
BlockBFT.update_forward_refs()
BlockPraos.update_forward_refs()
Metadata.update_forward_refs()
MetadataLabels1.update_forward_refs()
Metadatum4.update_forward_refs()
Metadatum5.update_forward_refs()
MetadatumMap.update_forward_refs()
Field14.update_forward_refs()
Script1.update_forward_refs()
ScriptNative2.update_forward_refs()
ScriptNative3.update_forward_refs()
ScriptNative4.update_forward_refs()
Transaction.update_forward_refs()



from .AcquireLedgerState import AcquireLedgerState
from .QueryBlockHeight import QueryBlockHeight
from .QueryNetworkTip import QueryNetworkTip
from .QueryStartTime import QueryStartTime
from .QueryGenesisConfiguration import QueryGenesisConfiguration
from .QueryLedgerTip import QueryLedgerTip
from .QueryEpoch import QueryEpoch
from .QueryEraStart import QueryEraStart
from .QueryEraSummaries import QueryEraSummaries
from .QueryLiveStakeDistribution import QueryLiveStakeDistribution
from .QueryStakePools import QueryStakePools
from .QueryUtxo import QueryUtxo
from .QueryProjectedRewards import QueryProjectedRewards
from .QueryRewardsProvenance import QueryRewardsProvenance
from .QueryRewardAccountSummaries import QueryRewardAccountSummaries
from .QueryProtocolParameters import QueryProtocolParameters
from .QueryProposedProtocolParameters import QueryProposedProtocolParameters
from .QueryConstitution import QueryConstitution
from .QueryConstitutionalCommittee import QueryConstitutionalCommittee
from .QueryTreasuryAndReserves import QueryTreasuryAndReserves

__all__ = [
    "AcquireLedgerState",
    "QueryBlockHeight",
    "QueryNetworkTip",
    "QueryStartTime",
    "QueryGenesisConfiguration",
    "QueryLedgerTip",
    "QueryEpoch",
    "QueryEraStart",
    "QueryEraSummaries",
    "QueryLiveStakeDistribution",
    "QueryStakePools",
    "QueryUtxo",
    "QueryProjectedRewards",
    "QueryRewardsProvenance",
    "QueryRewardAccountSummaries",
    "QueryProtocolParameters",
    "QueryProposedProtocolParameters",
    "QueryConstitution",
    "QueryConstitutionalCommittee",
    "QueryTreasuryAndReserves",
]



import pytest
import pycardano as pyc

from ogmios.errors import ResponseError
from tests.ogmios.test_fixtures import (
    client,
    chain_context,
    test_psk,
    test_ssk,
    test_pvk,
    test_svk,
    test_address,
)

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


@pytest.mark.skip
def test_EvaluateTransaction_empty(
    client, chain_context, test_psk, test_pvk, test_svk, test_address
):
    tx_builder = pyc.TransactionBuilder(chain_context)
    tx_builder.add_input_address(test_address)
    tx_builder.add_output(
        pyc.TransactionOutput(
            test_address,
            pyc.Value.from_primitive([1_000_000]),
        )
    )
    signed_tx = tx_builder.build_and_sign([test_psk])

    resp = client.evaluate_transaction.execute(signed_tx.to_cbor_hex())
    assert resp == ([], None)



from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ogmios.client import Client

from ogmios.errors import InvalidMethodError, InvalidResponseError, ResponseError
from ogmios.logger import logger
from ogmios.datatypes import Origin, Point, Tip
import ogmios.response_handler as rh
import ogmios.model.ogmios_model as om
import ogmios.model.model_map as mm

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


class FindIntersection:
    """Ogmios method to find a point on the blockchain. The first point that is found in the
    provided list will be returned.

    NOTE: This class is not intended to be used directly. Instead, use the Client.find_intersection
    method.

    :param client: The client to use for the request.
    :type client: Client
    """

    def __init__(self, client: Client):
        self.client = client
        self.method = mm.Method.findIntersection.value

    def execute(
        self, points: Union[List[Point, Origin]], id: Optional[Any] = None
    ) -> Tuple[Union[Point, Origin], Union[Tip, Origin], Optional[Any]]:
        """Send and receive the request.

        :param points: The list of points to find.
        :type points: list[Point | Origin]
        :param id: The ID of the request.
        :type id: Any
        :return: The intersection, tip, and ID of the response.
        :rtype: (Point | Origin, Tip | Origin, Optional[Any])
        """
        self.send(points, id)
        return self.receive()

    def send(self, points: List[Point, Origin], id: Optional[Any] = None) -> None:
        """Send the request.

        :param points: The list of points to find.
        :type points: list[Point | Origin]
        :param id: The ID of the request.
        :type id: Any
        """
        params = om.Params(points=[point._schematype for point in points])
        pld = om.FindIntersection(
            jsonrpc=self.client.rpc_version,
            method=self.method,
            params=params,
            id=id,
        )
        self.client.send(pld.json())

    def receive(self) -> Tuple[Union[Point, Origin], Union[Tip, Origin], Optional[Any]]:
        """Receive the response.

        :return: The intersection, tip, and ID of the response.
        :rtype: (Point | Origin, Tip | Origin, Optional[Any])
        """
        response = self.client.receive()
        return self._parse_FindIntersection_response(response)

    @staticmethod
    def _parse_FindIntersection_response(
        response: dict,
    ) -> Tuple[Union[Point, Origin], Union[Tip, Origin], Optional[Any]]:
        if response.get("method") != mm.Method.findIntersection.value:
            raise InvalidMethodError(f"Incorrect method for find_intersection response: {response}")

        if response.get("error"):
            raise ResponseError(f"Ogmios responded with error: {response}")

        # Successful response will contain intersection, tip, and id
        if result := response.get("result"):
            if (intersection_resp := result.get("intersection")) is not None and (
                tip_resp := result.get("tip")
            ) is not None:
                intersection: Union[Point, Origin] = rh.parse_PointOrOrigin(intersection_resp)
                tip: Union[Tip, Origin] = rh.parse_TipOrOrigin(tip_resp)
                id: Optional[Any] = response.get("id")
                logger.info(
                    f"""Parsed find_intersection response:
        Point = {intersection}
        Tip = {tip}
        ID = {id}"""
                )
                return intersection, tip, id

        raise InvalidResponseError(f"Failed to parse find_intersection response: {response}")



import ogmios

"""
This example prints the first 14 blocks of the Shelley era.

Set NETWORK to 'mainnet' or 'preprod' depending on the network Ogmios is connected to.
"""

NETWORK = "preprod"

if NETWORK == "mainnet":
    last_byron_block = ogmios.Point(
        slot=4492799, id="f8084c61b6a238acec985b59310b6ecec49c0ab8352249afd7268da5cff2a457"
    )
elif NETWORK == "preprod":
    last_byron_block = ogmios.Point(
        slot=84242, id="45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
    )
else:
    print("Invalid network. Please set NETWORK to 'mainnet' or 'preprod'.")
    exit(1)


def print_first_shelley_blocks():
    with ogmios.Client() as client:
        # Set chain pointer to origin
        try:
            point, tip, id = client.find_intersection.execute([last_byron_block])
        except ogmios.ResponseError:
            print("Intersection not found. Make sure you're connected to the proper network.")
            return

        blocks_printed = 0
        while True:
            direction, tip, point, id = client.next_block.execute()
            if direction.value == "forward":
                blocks_printed += 1
                print(f"Shelley block #{blocks_printed}: {point}")
                if blocks_printed >= 14:
                    break


if __name__ == "__main__":
    print_first_shelley_blocks()



from __future__ import annotations

from typing import Any, Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ogmios.client import Client

from ogmios.errors import InvalidMethodError, InvalidResponseError, ResponseError
from ogmios.logger import logger
from ogmios.datatypes import GenesisConfiguration, Era
import ogmios.model.ogmios_model as om
import ogmios.model.model_map as mm

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


class QueryGenesisConfiguration:
    """Ogmios method to query the genesis configuration of a specific era.

    NOTE: This class is not intended to be used directly. Instead, use the
    Client.query_genesis_configuration method.

    :param client: The client to use for the request.
    :type client: Client
    """

    def __init__(self, client: Client):
        self.client = client
        self.method = mm.Method.queryNetwork_genesisConfiguration.value

    def execute(
        self, era: mm.Era, id: Optional[Any] = None
    ) -> (GenesisConfiguration, Optional[Any]):
        """Send and receive the request.

        :param era: The era at which to query the genesis configuration.
        :type era: mm.Era
        :param id: The ID of the request.
        :type id: Any
        :return: The GenesisConfiguration at the specified era and ID of the response.
        :rtype: (GenesisConfiguration, Optional[Any])
        """
        self.send(era, id)
        return self.receive()

    def send(self, era: mm.Era, id: Optional[Any] = None) -> None:
        """Send the request.

        :param era: The era at which to query the genesis configuration.
        :type era: mm.Era
        :param id: The ID of the request.
        :type id: Any
        """
        params = om.Params9(era=era)
        pld = om.QueryNetworkGenesisConfiguration(
            jsonrpc=self.client.rpc_version,
            method=self.method,
            params=params,
            id=id,
        )
        self.client.send(pld.json())

    def receive(self) -> (GenesisConfiguration, Optional[Any]):
        """Receive a previously requested response.

        :return: The GenesisConfiguration at the specified era and ID of the response.
        :rtype: (GenesisConfiguration, Optional[Any])
        """
        response = self.client.receive()
        return self._parse_QueryGenesisConfiguration_response(response)

    @staticmethod
    def _parse_QueryGenesisConfiguration_response(
        response: dict,
    ) -> (GenesisConfiguration, Optional[Any]):
        if response.get("method") != mm.Method.queryNetwork_genesisConfiguration.value:
            raise InvalidMethodError(
                f"Incorrect method for query_genesis_configuration response: {response}"
            )

        if response.get("error"):
            raise ResponseError(f"Ogmios responded with error: {response}")

        # Successful response will contain success, point or origin, and ID
        if result := response.get("result"):
            genesis_configuration = None
            if (era := result.get("era")) is not None:
                if (
                    era == Era.byron.value
                    and (genesis_key_hashes := result.get("genesisKeyHashes")) is not None
                    and (genesis_delegations := result.get("genesisDelegations")) is not None
                    and (start_time := result.get("startTime")) is not None
                    and (initial_funds := result.get("initialFunds")) is not None
                    and (initial_vouchers := result.get("initialVouchers")) is not None
                    and (security_parameter := result.get("securityParameter")) is not None
                    and (network_magic := result.get("networkMagic")) is not None
                    and (updatable_parameters := result.get("updatableParameters")) is not None
                ):
                    genesis_configuration = GenesisConfiguration(
                        era=era,
                        genesis_key_hashes=genesis_key_hashes,
                        genesis_delegations=genesis_delegations,
                        start_time=start_time,
                        initial_funds=initial_funds,
                        initial_vouchers=initial_vouchers,
                        security_parameter=security_parameter,
                        network_magic=network_magic,
                        updatable_parameters=updatable_parameters,
                    )
                elif (
                    era == Era.shelley.value
                    and (start_time := result.get("startTime")) is not None
                    and (network_magic := result.get("networkMagic")) is not None
                    and (network := result.get("network")) is not None
                    and (active_slots_coefficient := result.get("activeSlotsCoefficient"))
                    is not None
                    and (security_parameter := result.get("securityParameter")) is not None
                    and (epoch_length := result.get("epochLength")) is not None
                    and (slots_per_kes_period := result.get("slotsPerKesPeriod")) is not None
                    and (max_kes_evolutions := result.get("maxKesEvolutions")) is not None
                    and (slot_length := result.get("slotLength")) is not None
                    and (update_quorum := result.get("updateQuorum")) is not None
                    and (max_lovelace_supply := result.get("maxLovelaceSupply")) is not None
                    and (initial_parameters := result.get("initialParameters")) is not None
                    and (initial_delegates := result.get("initialDelegates")) is not None
                    and (initial_funds := result.get("initialFunds")) is not None
                    and (initial_stake_pools := result.get("initialStakePools")) is not None
                ):
                    genesis_configuration = GenesisConfiguration(
                        era=era,
                        start_time=start_time,
                        network_magic=network_magic,
                        network=network,
                        active_slots_coefficient=active_slots_coefficient,
                        security_parameter=security_parameter,
                        epoch_length=epoch_length,
                        slots_per_kes_period=slots_per_kes_period,
                        max_kes_evolutions=max_kes_evolutions,
                        slot_length=slot_length,
                        update_quorum=update_quorum,
                        max_lovelace_supply=max_lovelace_supply,
                        initial_parameters=initial_parameters,
                        initial_delegates=initial_delegates,
                        initial_funds=initial_funds,
                        initial_stake_pools=initial_stake_pools,
                    )
                elif (
                    era == Era.alonzo.value
                    and (updatable_parameters := result.get("updatableParameters")) is not None
                ):
                    genesis_configuration = GenesisConfiguration(
                        era=era, updatableParameters=updatable_parameters
                    )
                elif (
                    era == Era.conway.value
                    and (constitution := result.get("constitution")) is not None
                    and (constitutional_committee := result.get("constitutionalCommittee"))
                    is not None
                    and (updatable_parameters := result.get("updatableParameters")) is not None
                ):
                    genesis_configuration = GenesisConfiguration(
                        era=era,
                        constitution=constitution,
                        constitutional_committee=constitutional_committee,
                        updatableParameters=updatable_parameters,
                    )

            if genesis_configuration:
                id: Optional[Any] = response.get("id")
                logger.info(
                    f"""Parsed query_genesis_configuration response:
        Configuration = {genesis_configuration}
        ID = {id}"""
                )
                return genesis_configuration, id
        raise InvalidResponseError(
            f"Failed to parse query_genesis_configuration response: {response}"
        )



import pytest
import pycardano as pyc

from ogmios import Point
from ogmios.client import Client
from ogmios import utils


@pytest.fixture
def client():
    with Client() as client:
        yield client


@pytest.fixture
def chain_context():
    yield pyc.OgmiosV6ChainContext("localhost", 1337)


@pytest.fixture
def era(client):
    yield utils.get_current_era(client)


@pytest.fixture
def test_psk():
    yield pyc.PaymentExtendedSigningKey.load("./tests/test_wallet/test_addr0.skey")


@pytest.fixture
def test_ssk():
    yield pyc.StakeExtendedSigningKey.load("./tests/test_wallet/test_stake.skey")


@pytest.fixture
def test_pvk(test_psk):
    yield pyc.PaymentVerificationKey.from_signing_key(test_psk)


@pytest.fixture
def test_svk(test_ssk):
    yield pyc.StakeVerificationKey.from_signing_key(test_ssk)


@pytest.fixture
def test_address(test_pvk, test_svk):
    yield pyc.Address(test_pvk.hash(), test_svk.hash(), pyc.Network.TESTNET)


@pytest.fixture
def last_byron_block_point():
    yield Point(slot=84242, id="45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af")


@pytest.fixture
def last_shelley_block_point():
    yield Point(slot=518360, id="f9d8b6c77fedd60c3caf5de0ce63a0aeb9d1753269c9c07503d9aa09d5144481")


@pytest.fixture
def last_allegra_block_point():
    yield Point(slot=950340, id="74c03af754bcde9cd242c5a168689edcab1756a3f7ae4d5dca1a31d86839c7b1")


@pytest.fixture
def last_mary_block_point():
    yield Point(slot=1382348, id="af5fddc7d16a349e1a2af8ba89f4f5d3273955a13095b3709ef6e3db576a0b33")


@pytest.fixture
def last_alonzo_block_point():
    yield Point(slot=3542390, id="f93e682d5b91a94d8660e748aef229c19cb285bfb9830db48941d6a78183d81f")



import pytest

from tests.ogmios.test_fixtures import client, era
from ogmios.datatypes import Era


def test_QueryConstitution(client, era):
    if Era.get_index(era) < Era.get_index(Era.conway):
        pytest.skip("Test only valid for conway era")

    id_str = "My ID string"
    guardrails, metadata, id = client.query_constitution.execute(id_str)

    if guardrails:
        assert isinstance(guardrails, dict)
        assert guardrails.keys() == {"hash"}
    if metadata:
        assert isinstance(metadata, dict)
        assert metadata.keys() == {"url", "hash"}
    assert id == id_str



import ogmios

"""
This example prints all transactions involving a specific wallet address.
"""

target_addr = "addr_test1vzpwq95z3xyum8vqndgdd9mdnmafh3djcxnc6jemlgdmswcve6tkw"
batch_size = 1000


def print_first_shelley_blocks():
    with ogmios.Client() as client:
        # Set chain pointer to origin
        point, _, _ = client.find_intersection.execute([ogmios.Origin()])

        txs_found = 0
        while True:
            # Batch requests to improve performance
            for i in range(batch_size):
                client.next_block.send()

            for i in range(batch_size):
                direction, tip, block, id = client.next_block.receive()
                if direction.value == "forward":
                    # Find transactions involving the target address
                    if isinstance(block, ogmios.Block) and hasattr(block, "transactions"):
                        for tx in block.transactions:
                            if tx.get("outputs"):
                                for output in tx["outputs"]:
                                    if output["address"] == target_addr:
                                        txs_found += 1
                                        print(f"Transaction #{txs_found}: {tx.get('id')}")
                                        break

                    # Stop when we've reached the network tip
                    if tip.height == block.height:
                        print(f"Reached chain tip at slot {tip.slot}")
                        return


if __name__ == "__main__":
    print_first_shelley_blocks()



from pycardano import TransactionBody

import ogmios
import pycardano as pyc

"""
This example gets the contents of the mempool and waits for it to be empty.
Note: for this example you need to install pycardano
"""


def submit_transaction():
    """Submit a transaction so we have something in the mempool."""
    with pyc.OgmiosV6ChainContext("localhost", 1337) as context:
        network = pyc.Network.TESTNET
        psk = pyc.PaymentExtendedSigningKey.load("./tests/test_wallet/test_addr0.skey")
        ssk = pyc.StakeExtendedSigningKey.load("./tests/test_wallet/test_stake.skey")
        pvk = pyc.PaymentVerificationKey.from_signing_key(psk)
        svk = pyc.StakeVerificationKey.from_signing_key(ssk)
        address = pyc.Address(pvk.hash(), svk.hash(), network)
        builder = pyc.TransactionBuilder(context)

        builder.add_input_address(address)
        builder.add_output(
            pyc.TransactionOutput(
                address,
                pyc.Value.from_primitive(
                    [
                        10000000,
                    ]
                ),
            )
        )
        signed_tx = builder.build_and_sign([psk], change_address=address)
        context.submit_tx(signed_tx)


def tx_dict_to_pycardano_tx(tx_dict: dict) -> TransactionBody:
    """
    Convert a transaction dictionary to a PyCardano Transaction

    :param tx_dict: The transaction dictionary
    :type tx_dict: dict
    :return: The PyCardano Transaction
    :rtype: pyc.Transaction
    """
    return TransactionBody(tx_dict)


def get_mempool_contents():
    """Get the contents of the mempool and wait for it to be empty."""
    with ogmios.Client() as client:
        mempool_txs = ogmios.utils.get_mempool_transactions(client)
        print(f"Mempool has {len(mempool_txs)} transactions:\n{mempool_txs}")

        # Optionally, turn the TX dicts into PyCardano TransactionBody objects
        for tx in mempool_txs:
            pyc_tx = ogmios.utils.tx_dict_to_pycardano_tx(tx)

        print("Waiting for mempool to be empty...")
        ogmios.utils.wait_for_empty_mempool(client)
        print("Mempool is empty!")


if __name__ == "__main__":
    submit_transaction()
    get_mempool_contents()



import pytest

from ogmios.errors import ResponseError
from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_NextTransaction(client):
    id_str = "My ID string"
    with pytest.raises(ResponseError):
        # Try getting mempool size before acquiring a snapshot
        client.next_transaction.execute()

    # Acquire a mempool snapshot
    client.acquire_mempool.execute()
    _, id = client.next_transaction.execute(id_str)
    assert id == id_str



from ogmios import ProtocolParameters
from tests.ogmios.test_fixtures import client


def test_QueryProtocolParameters(client):
    id_str = "My ID string"
    result, id = client.query_protocol_parameters.execute(id_str)

    assert isinstance(result, ProtocolParameters)
    assert id == id_str



import pytest

from ogmios.datatypes import Origin, Direction, Era

from tests.ogmios.test_fixtures import (
    client,
    last_byron_block_point,
    last_shelley_block_point,
    last_allegra_block_point,
    last_mary_block_point,
    last_alonzo_block_point,
    era,
)

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_NextBlock_origin(client):
    intersection, _, _ = client.find_intersection.execute([Origin()])
    assert isinstance(intersection, Origin)

    my_id = "12345"
    direction, tip, block, id = client.next_block.execute(my_id)
    assert direction == Direction.backward
    assert block == Origin()
    assert id == my_id
    assert tip.height > 0
    assert tip.slot > 0

    direction, _, block, _ = client.next_block.execute()
    assert direction == Direction.forward
    assert block.height == 0
    assert block.era == "byron"

    direction, _, block, _ = client.next_block.execute()
    assert direction == Direction.forward
    assert block.height == 1
    assert block.era == "byron"


def test_NextBlock_byronEra(client):
    intersection, _, _ = client.find_intersection.execute([Origin()])
    assert isinstance(intersection, Origin)

    for i in range(10):
        direction, tip, block, _ = client.next_block.execute()
        if direction.value == "forward":
            assert block.era == Era.byron.value


def test_NextBlock_shelleyEra(client, era, last_byron_block_point):
    if Era.get_index(era) < Era.get_index(Era.shelley):
        pytest.skip("Test only valid for shelley era")

    intersection, _, _ = client.find_intersection.execute([last_byron_block_point, Origin()])

    for _ in range(10):
        direction, tip, block, _ = client.next_block.execute()
        if direction.value == "forward":
            print(f"Slot: {block.slot}, Hash: {block.id}, Era: {block.era}")
            assert block.era == Era.shelley.value


def test_NextBlock_allegraEra(client, era, last_shelley_block_point):
    if Era.get_index(era) < Era.get_index(Era.allegra):
        pytest.skip("Test only valid for allegra era")

    intersection, _, _ = client.find_intersection.execute([last_shelley_block_point, Origin()])

    for _ in range(10):
        direction, tip, block, _ = client.next_block.execute()
        if direction.value == "forward":
            print(f"Slot: {block.slot}, Hash: {block.id}, Era: {block.era}")
            assert block.era == Era.allegra.value


def test_NextBlock_maryEra(client, era, last_allegra_block_point):
    if Era.get_index(era) < Era.get_index(Era.mary):
        pytest.skip("Test only valid for mary era")

    intersection, _, _ = client.find_intersection.execute([last_allegra_block_point, Origin()])

    for _ in range(10):
        direction, tip, block, _ = client.next_block.execute()
        if direction.value == "forward":
            print(f"Slot: {block.slot}, Hash: {block.id}, Era: {block.era}")
            assert block.era == Era.mary.value


def test_NextBlock_alonzoEra(client, era, last_mary_block_point):
    if Era.get_index(era) < Era.get_index(Era.alonzo):
        pytest.skip("Test only valid for alonzo era")

    intersection, _, _ = client.find_intersection.execute([last_mary_block_point, Origin()])

    for _ in range(10):
        direction, tip, block, _ = client.next_block.execute()
        if direction.value == "forward":
            print(f"Slot: {block.slot}, Hash: {block.id}, Era: {block.era}")
            assert block.era == Era.alonzo.value


def test_NextBlock_babbageEra(client, era, last_alonzo_block_point):
    if Era.get_index(era) < Era.get_index(Era.babbage):
        pytest.skip("Test only valid for babbage era")

    intersection, _, _ = client.find_intersection.execute([last_alonzo_block_point, Origin()])

    for _ in range(10):
        direction, tip, block, _ = client.next_block.execute()
        if direction.value == "forward":
            print(f"Slot: {block.slot}, Hash: {block.id}, Era: {block.era}")
            assert block.era == Era.babbage.value



"""This module contains user-friendly datatypes for Ogmios objects.

    Behind the scenes, ogmios-python uses objects derived from the cardano.json and ogmios.json
    schema files provided by Ogmios. These types are useful for validating that data passed to
    and from the Ogmios server is properly formatted. However, the schema types are not
    conducive to Pythonic use. Therefore, we use intermediate datatypes for user-facing
    functions of this library.

    .. seealso::
        Cardano schema: `cardano.json <https://ogmios.dev/cardano.json>`_,
        Ogmios schema: `ogmios.json schema <https://ogmios.dev/ogmios.json>`_

"""

from datetime import datetime
from enum import Enum
from typing import Dict, Union
from fractions import Fraction

from ogmios.errors import InvalidOgmiosParameter
import ogmios.model.cardano_model as cm
import ogmios.model.ogmios_model as om
import ogmios.model.model_map as mm

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


class Era(Enum):
    """An enum representing the eras of the Cardano blockchain."""

    byron = "byron"
    shelley = "shelley"
    allegra = "allegra"
    mary = "mary"
    alonzo = "alonzo"
    babbage = "babbage"
    conway = "conway"

    @staticmethod
    def by_index(index):
        return list(Era)[index]

    @staticmethod
    def get_index(era):
        return list(Era).index(era)

    @staticmethod
    def is_genesis_era(era):
        return era in [Era.byron, Era.shelley, Era.alonzo, Era.conway]


class Direction(Enum):
    """An enum representing the direction of a blockchain traversal."""

    forward = "forward"
    backward = "backward"


class Origin:
    """A class representing the origin of the blockchain.

    The origin is the first block in the blockchain. It is the only block that does not have a
    parent block.
    """

    def __init__(self):
        self._schematype = om.Origin.origin

    def __eq__(self, other):
        return True if isinstance(other, Origin) else False

    def __str__(self):
        return "Origin()"


class Point:
    """A class representing a point in the blockchain.

    :param slot: The slot number of the point.
    :type slot: int
    :param id: The block hash of the point.
    :type id: str
    """

    def __init__(self, slot: int, id: str):
        self.slot = slot
        self.id = id
        self._schematype = om.PointOrOrigin1(slot=self.slot, id=self.id)

    def __eq__(self, other):
        if isinstance(other, Point):
            if self.slot == other.slot and self.id == other.id:
                return True
        return False

    def __str__(self):
        return f"Point(Slot={self.slot:,}, ID={self.id})"


class Tip:
    """A class representing the tip of the blockchain.

    :param slot: The slot number of the tip.
    :type slot: int
    :param id: The block hash of the tip.
    :type id: str
    :param height: The block height of the tip.
    :type height: int
    """

    def __init__(self, slot: int, id: str, height: int):
        self.slot = slot
        self.id = id
        self.height = height
        self._schematype = om.Tip(slot=self.slot, id=self.id, height=self.height)

    def __eq__(self, other):
        if isinstance(other, Tip):
            if self.slot == other.slot and self.id == other.id and self.height == other.height:
                return True
        return False

    def __str__(self):
        return f"Tip(Slot={self.slot:,}, Height={self.height:,}, ID={self.id})"

    def to_point(self) -> Point:
        """Returns a Point representation of the Tip"""
        return Point(self.slot, self.id)


class Block:
    """
    Represents a block in the blockchain.

    :param blocktype: The type of the block (EBB, BFT, or Praos)
    :type blocktype: ogmios.model.model_map.Types
    :param kwargs: Additional arguments depending on the block type.
    :raises InvalidOgmiosParameter: If an unsupported block type is provided.
    """

    def __init__(self, blocktype: mm.Types, **kwargs):
        self.blocktype = blocktype
        if blocktype == mm.Types.ebb.value:
            self.era = kwargs.get("era")
            self.id = kwargs.get("id")
            self.ancestor = kwargs.get("ancestor")
            self.height = kwargs.get("height")
            self._schematype = cm.BlockEBB(
                type=self.blocktype,
                era=self.era,
                id=self.id,
                ancestor=self.ancestor,
                height=self.height,
            )
        elif blocktype == mm.Types.bft.value:
            self.era = kwargs.get("era")
            self.id = kwargs.get("id")
            self.ancestor = kwargs.get("ancestor")
            self.height = kwargs.get("height")
            self.slot = kwargs.get("slot")
            self.size = kwargs.get("size")
            self.transactions = kwargs.get("transactions")
            self.protocol = kwargs.get("protocol")
            self.issuer = kwargs.get("issuer")
            self.delegate = kwargs.get("delegate")
            self._schematype = cm.BlockBFT(
                type=self.blocktype,
                era=self.era,
                id=self.id,
                ancestor=self.ancestor,
                height=self.height,
                slot=self.slot,
                size=self.size,
                transactions=self.transactions,
                protocol=self.protocol,
                issuer=self.issuer,
                delegate=self.delegate,
            )
        elif blocktype == mm.Types.praos.value:
            self.era = kwargs.get("era")
            self.id = kwargs.get("id")
            self.ancestor = kwargs.get("ancestor")
            self.nonce = kwargs.get("nonce")
            self.height = kwargs.get("height")
            self.slot = kwargs.get("slot")
            self.size = kwargs.get("size")
            self.transactions = kwargs.get("transactions")
            self.protocol = kwargs.get("protocol")
            self.issuer = kwargs.get("issuer")
            self._schematype = cm.BlockPraos(
                type=self.blocktype,
                era=self.era,
                id=self.id,
                ancestor=self.ancestor,
                height=self.height,
                slot=self.slot,
                size=self.size,
                protocol=self.protocol,
                issuer=self.issuer,
            )
        else:
            raise InvalidOgmiosParameter(f"Unsupported block type: {blocktype}")

    def __str__(self):
        """
        Returns a string representation of the block.

        :return: A string representation of the block.
        :rtype: str
        """
        if self.blocktype == mm.Types.ebb.value:
            return f"Block(Type=EBB, Era={self.era}, ID={self.id}, Ancestor={self.ancestor}, Height={self.height:,})"
        elif self.blocktype == mm.Types.bft.value:
            return f"Block(Type=BFT, Era={self.era}, ID={self.id}, Ancestor={self.ancestor}, Height={self.height:,}, Slot={self.slot:,}, Size={self.size.get('bytes'):,}, TXs={len(self.transactions)})"
        elif self.blocktype == mm.Types.praos.value:
            return f"Block(Type=Praos, Era={self.era}, ID={self.id}, Ancestor={self.ancestor}, Height={self.height:,}, Slot={self.slot:,}, Size={self.size.get('bytes'):,}, TXs={len(self.transactions)})"

    def __eq__(self, other):
        if isinstance(other, Block):
            if (
                self.blocktype == other.blocktype
                and self.era == other.era
                and self.id == other.id
                and self.ancestor == other.ancestor
                and self.height == other.height
                and self.slot == other.slot
                and self.size == other.size
                and self.transactions == other.transactions
                and self.protocol == other.protocol
                and self.issuer == other.issuer
                and self.delegate == other.delegate
            ):
                return True
        return False

    def to_point(self) -> Point:
        """Returns a Point representation of the Block"""
        return Point(self.slot, self.id)

    def to_tip(self) -> Tip:
        """Returns a Tip representation of the Block"""
        return Tip(self.slot, self.id, self.height)


class Script:
    """A class representing a script.

    :param language: The language of the script.
    :type language: str
    :param cbor: The CBOR representation of the script.
    :type cbor: str
    :param json: The JSON representation of the script.
    :type json: dict
    """

    def __init__(
        self, language: str, cbor: Union[str, None] = None, json: Union[Dict, None] = None
    ):
        self.language = language
        self.cbor = cbor
        self.json = json
        if language == "native":
            self._schematype = om.Script1(language=self.language, cbor=self.cbor, json=self.json)
        else:
            self._schematype = om.Script2(language=self.language, cbor=self.cbor)


class Utxo:
    """A class representing a UTxO.

    :param tx_id: The transaction ID of the UTxO.
    :type tx_id: str
    :param index: The index of the UTxO.
    :type index: int
    :param address: The address of the UTxO.
    :type address: str
    :param value: The lovelace value of the UTxO (ada / 1e6)
    :type value: int
    :param datum_hash: The datum hash of the UTxO.
    :type datum_hash: str
    :param datum: The datum of the UTxO.
    :type datum: str
    :param script: The script of the UTxO.
    :type script: Script
    """

    def __init__(
        self,
        tx_id: str,
        index: int,
        address: str,
        value: dict,
        datum_hash: Union[str, None] = None,
        datum: Union[str, None] = None,
        script: Union[Script, None] = None,
    ):
        self.tx_id = tx_id
        self.index = index
        self.address = address
        self.value = value
        self.datum_hash = datum_hash
        self.datum = datum
        self.script = script

        # Not sure why the model Utxo class is a list of UtxoItems
        self._schematype = om.Utxo(
            __root__=[
                om.UtxoItem(
                    transaction=om.Transaction5(id=self.tx_id),
                    index=self.index,
                    address=om.Address(__root__=self.address),
                    value=om.Value(ada=om.Ada(lovelace=self.value.get("ada").get("lovelace"))),
                    datumHash=self.datum_hash,
                    datum=self.datum,
                    script=self.script,
                )
            ]
        )

    def __eq__(self, other):
        if isinstance(other, Utxo):
            if (
                self.tx_id == other.tx_id
                and self.index == other.index
                and self.address == other.address
                and self.value == other.value
                and self.datum_hash == other.datum_hash
                and self.datum == other.datum
                and self.script == other.script
            ):
                return True
        return False

    def __str__(self):
        return f"Utxo(TX={self.tx_id}, Index={self.index}, Address={self.address}, Value={self.value.get('ada'):,})"


class TxOutputReference:
    """A class representing a transaction output reference.

    :param transaction: The transaction ID of the output.
    :type transaction: str
    :param index: The index of the output.
    :type index: int
    """

    def __init__(self, tx_id: str, index: int):
        self.tx_id = tx_id
        self.index = index
        self._schematype = om.TransactionOutputReference(
            transaction={"id": self.tx_id}, index=self.index
        )


class Address:
    """A class representing an address."""

    def __init__(self, address: str):
        self.address = address
        self._schematype = om.Address(__root__=self.address)


class GenesisConfiguration:
    """A class representing the genesis configuration of the blockchain. Input parameters depend
    on the era of the genesis configuration.

    :raises InvalidOgmiosParameter: If an unsupported era is provided.
    """

    def __init__(
        self,
        era: Era,
        **kwargs,
    ):
        if era == Era.byron.value:
            self.era = era
            self.genesis_key_hashes = kwargs.get("genesis_key_hashes")
            self.genesis_delegations = kwargs.get("genesis_delegations")
            self.start_time = datetime.strptime(kwargs.get("start_time"), "%Y-%m-%dT%H:%M:%SZ")
            self.initial_funds = kwargs.get("initial_funds")
            self.initial_vouchers = kwargs.get("initial_vouchers")
            self.security_parameter = kwargs.get("security_parameter")
            self.network_magic = kwargs.get("network_magic")
            self.updatable_parameters = (
                BootstrapProtocolParameters(**kwargs.get("updatable_parameters"))
                if kwargs.get("updatable_parameters")
                else None
            )
            self._schematype = om.GenesisByron(
                era=self.era,
                genesisKeyHashes=self.genesis_key_hashes,
                genesisDelegations=self.genesis_delegations,
                startTime=self.start_time,
                initialFunds=self.initial_funds,
                initialVouchers=self.initial_vouchers,
                securityParameter=self.security_parameter,
                networkMagic=self.network_magic,
                updatableParameters=(
                    self.updatable_parameters._schema_type if self.updatable_parameters else None
                ),
            )
        elif era == Era.shelley.value:
            self.era = era
            self.start_time = datetime.strptime(kwargs.get("start_time"), "%Y-%m-%dT%H:%M:%SZ")
            self.network_magic = kwargs.get("network_magic")
            self.network = kwargs.get("network")
            self.active_slots_coefficient = float(Fraction(kwargs.get("active_slots_coefficient")))
            self.security_parameter = kwargs.get("security_parameter")
            self.epoch_length = kwargs.get("epoch_length")
            self.slots_per_kes_period = kwargs.get("slots_per_kes_period")
            self.max_kes_evolutions = kwargs.get("max_kes_evolutions")
            self.slot_length = kwargs.get("slot_length").get("milliseconds")
            self.update_quorum = kwargs.get("update_quorum")
            self.max_lovelace_supply = kwargs.get("max_lovelace_supply")
            self.initial_parameters = ProtocolParameters(**kwargs.get("initial_parameters"))
            self.initial_delegates = kwargs.get("initial_delegates")
            self.initial_funds = kwargs.get("initial_funds")
            self.initial_stake_pools = kwargs.get("initial_stake_pools")
            self._schematype = om.GenesisShelley(
                era=self.era,
                startTime=self.start_time,
                networkMagic=self.network_magic,
                network=self.network,
                activeSlotsCoefficient=kwargs.get("active_slots_coefficient"),
                securityParameter=self.security_parameter,
                epochLength=self.epoch_length,
                slotsPerKesPeriod=self.slots_per_kes_period,
                maxKesEvolutions=self.max_kes_evolutions,
                slotLength=kwargs.get("slot_length"),
                updateQuorum=self.update_quorum,
                maxLovelaceSupply=self.max_lovelace_supply,
                initialParameters=self.initial_parameters._schema_type,
                initialDelegates=self.initial_delegates,
                initialFunds=self.initial_funds,
                initialStakePools=self.initial_stake_pools,
            )
        elif era == Era.alonzo.value:
            self.era = era
            self.updatable_parameters = AlonzoUpdatableParameters(
                **kwargs.get("updatableParameters")
            )
            self._schematype = om.GenesisAlonzo(
                era=self.era, updatableParameters=self.updatable_parameters._schema_type
            )
        elif era == Era.conway.value:
            self.era = era
            self.constitution = kwargs.get("constitution")
            self.constitutional_committee = kwargs.get("constitutional_committee")
            self.updatable_parameters = ConwayUpdatableParameters(
                **kwargs.get("updatableParameters")
            )
            self._schematype = om.GenesisConway(
                era=self.era,
                constitution=self.constitution,
                constitutionalCommittee=self.constitutional_committee,
                updatableParameters=self.updatable_parameters._schema_type,
            )
        else:
            raise InvalidOgmiosParameter(f"Unsupported era for GenesisConfiguration: {era}")

    def __eq__(self, other):
        if isinstance(other, GenesisConfiguration):
            if (
                self.era == other.era
                and self.genesis_key_hashes == other.genesis_key_hashes
                and self.genesis_delegations == other.genesis_delegations
                and self.start_time == other.start_time
                and self.initial_funds == other.initial_funds
                and self.initial_vouchers == other.initial_vouchers
                and self.security_parameter == other.security_parameter
                and self.network_magic == other.network_magic
                and self.updatable_parameters == other.updatable_parameters
            ):
                return True
        return False


class EraSummary:
    """Summary of the slotting parameters and boundaries for each known era. Useful for doing
    slot-arithmetic and time conversions.

    :param start_time: The start time of the era (in seconds, relative to the network start).
    :type start_time: int
    :param start_slot: The start slot of the era.
    :type start_slot: int
    :param start_epoch: The start epoch of the era.
    :type start_epoch: int
    :param end_time: The end time of the era (in seconds, relative to the network start).
    :type end_time: int
    :param end_slot: The end slot of the era.
    :type end_slot: int
    :param end_epoch: The end epoch of the era.
    :type end_epoch: int
    :param epoch_length: The epoch length of the era.
    :type epoch_length: int
    :param slot_length: The slot length of the era, in milliseconds.
    :type slot_length: int
    :param safe_zone: Number of slots from the tip of the ledger in which it is guaranteed
        that no hard fork can take place. This should be (at least) the number of slots in
        which we are guaranteed to have k blocks.
    :type safe_zone: int
    """

    def __init__(
        self,
        start_time: int,
        start_slot: int,
        start_epoch: int,
        end_time: int,
        end_slot: int,
        end_epoch: int,
        epoch_length: int,
        slot_length: int,
        safe_zone: Union[int, None] = None,
    ):
        self.start_time = start_time
        self.start_slot = start_slot
        self.start_epoch = start_epoch
        self.end_time = end_time
        self.end_slot = end_slot
        self.end_epoch = end_epoch
        self.epoch_length = epoch_length
        self.slot_length = slot_length
        self.safe_zone = safe_zone
        self._schematype = None

    def __str__(self):
        return f"EraSummary(Time={self.start_time:,}-{self.end_time:,} s, Epoch={self.start_epoch:,}-{self.end_epoch:,}, Slot={self.start_slot:,}-{self.end_slot:,})"


class Ada:
    """A class representing an amount of ada.

    :param ada: The amount of ada.
    :type ada: int
    """

    def __init__(self, amount: int, is_lovelace: bool = False):
        self.ada = amount / 1e6 if is_lovelace else amount
        self.lovelace = int(amount if is_lovelace else amount * 1e6)
        self._schematype = om.Ada(lovelace=self.lovelace)

    def __eq__(self, other):
        if isinstance(other, Ada):
            if self.ada == other.ada:
                return True
        return False

    def __str__(self):
        return f"{self.ada:,} ADA"

    def __dict__(self):
        return {"ada": {"lovelace": self.lovelace}}


class BootstrapProtocolParameters:
    """A class representing the bootstrap protocol parameters of the blockchain.

    :param heavyDelegationThreshold: The heavy delegation threshold.
    :type heavyDelegationThreshold: dict
    :param maxBlockBodySize: The maximum block body size.
    :type maxBlockBodySize: dict
    :param maxBlockHeaderSize: The maximum block header size.
    :type maxBlockHeaderSize: dict
    :param maxUpdateProposalSize: The maximum update proposal size.
    :type maxUpdateProposalSize: dict
    :param maxTransactionSize: The maximum transaction size.
    :type maxTransactionSize: dict
    :param multiPartyComputationThreshold: The multi-party computation threshold.
    :type multiPartyComputationThreshold: dict
    :param scriptVersion: The script version.
    :type scriptVersion: int
    :param slotDuration: The slot duration.
    :type slotDuration: int
    :param unlockStakeEpoch: The unlock stake epoch.
    :type unlockStakeEpoch: int
    :param updateProposalThreshold: The update proposal threshold.
    :type updateProposalThreshold: dict
    :param updateProposalTimeToLive: The update proposal time to live.
    :type updateProposalTimeToLive: int
    :param updateVoteThreshold: The update vote threshold.
    :type updateVoteThreshold: dict
    :param softForkInitThreshold: The soft fork init threshold.
    :type softForkInitThreshold: dict
    :param softForkMinThreshold: The soft fork min threshold.
    :type softForkMinThreshold: dict
    :param softForkDecrementThreshold: The soft fork decrement threshold.
    :type softForkDecrementThreshold: dict
    :param minFeeCoefficient: The minimum fee coefficient.
    :type minFeeCoefficient: int
    :param minFeeConstant: The minimum fee constant.
    :type minFeeConstant: Lovelace
    """

    def __init__(
        self,
        heavyDelegationThreshold: dict = None,
        maxBlockBodySize: dict = None,
        maxBlockHeaderSize: dict = None,
        maxUpdateProposalSize: dict = None,
        maxTransactionSize: dict = None,
        multiPartyComputationThreshold: dict = None,
        scriptVersion: int = None,
        slotDuration: int = None,
        unlockStakeEpoch: int = None,
        updateProposalThreshold: dict = None,
        updateProposalTimeToLive: int = None,
        updateVoteThreshold: dict = None,
        softForkInitThreshold: dict = None,
        softForkMinThreshold: dict = None,
        softForkDecrementThreshold: dict = None,
        minFeeCoefficient: int = None,
        minFeeConstant: Ada = None,
    ):
        self.heavy_delegation_threshold = heavyDelegationThreshold
        self.max_block_body_size = maxBlockBodySize
        self.max_block_header_size = maxBlockHeaderSize
        self.max_update_proposal_size = maxUpdateProposalSize
        self.max_transaction_size = maxTransactionSize
        self.multi_party_computation_threshold = multiPartyComputationThreshold
        self.script_version = scriptVersion
        self.slot_duration = slotDuration
        self.unlock_stake_epoch = unlockStakeEpoch
        self.update_proposal_threshold = updateProposalThreshold
        self.update_proposal_time_to_live = updateProposalTimeToLive
        self.update_vote_threshold = updateVoteThreshold
        self.soft_fork_init_threshold = softForkInitThreshold
        self.soft_fork_min_threshold = softForkMinThreshold
        self.soft_fork_decrement_threshold = softForkDecrementThreshold
        self.min_fee_coefficient = minFeeCoefficient
        self.min_fee_constant = minFeeConstant
        self._schema_type = om.BootstrapProtocolParameters(
            heavyDelegationThreshold=self.heavy_delegation_threshold,
            maxBlockBodySize=self.max_block_body_size,
            maxBlockHeaderSize=self.max_block_header_size,
            maxUpdateProposalSize=self.max_update_proposal_size,
            maxTransactionSize=self.max_transaction_size,
            multiPartyComputationThreshold=self.multi_party_computation_threshold,
            scriptVersion=self.script_version,
            slotDuration=self.slot_duration,
            unlockStakeEpoch=self.unlock_stake_epoch,
            updateProposalThreshold=self.update_proposal_threshold,
            updateProposalTimeToLive=self.update_proposal_time_to_live,
            updateVoteThreshold=self.update_vote_threshold,
            softForkInitThreshold=self.soft_fork_init_threshold,
            softForkMinThreshold=self.soft_fork_min_threshold,
            softForkDecrementThreshold=self.soft_fork_decrement_threshold,
            minFeeCoefficient=self.min_fee_coefficient,
            minFeeConstant=self.min_fee_constant,
        )

    def _to_ada(self, value: Union[int, Dict, Ada]) -> Ada:
        if isinstance(value, Ada):
            return value
        elif isinstance(value, dict):
            return Ada(value.get("ada").get("lovelace"), is_lovelace=True)
        elif isinstance(value, int):
            return Ada(value)
        else:
            raise InvalidOgmiosParameter(f"Invalid type for value {value}: {type(value)}")


class ProtocolParameters:
    """A class representing the protocol parameters of the blockchain.

    :param minFeeCoefficient: The minimum fee coefficient.
    :type minFeeCoefficient: int
    :param minFeeConstant: The minimum fee constant.
    :type minFeeConstant: Ada
    :param minUtxoDepositCoefficient: The minimum UTXO deposit coefficient.
    :type minUtxoDepositCoefficient: int
    :param minUtxoDepositConstant: The minimum UTXO deposit constant.
    :type minUtxoDepositConstant: Ada
    :param maxBlockBodySize: The maximum block body size.
    :type maxBlockBodySize: dict
    :param maxBlockHeaderSize: The maximum block header size.
    :type maxBlockHeaderSize: dict
    :param stakeCredentialDeposit: The stake credential deposit.
    :type stakeCredentialDeposit: Ada
    :param stakePoolDeposit: The stake pool deposit.
    :type stakePoolDeposit: Ada
    :param stakePoolRetirementEpochBound: The stake pool retirement epoch bound.
    :type stakePoolRetirementEpochBound: int
    :param stakePoolPledgeInfluence: The stake pool pledge influence.
    :type stakePoolPledgeInfluence: str
    :param minStakePoolCost: The minimum stake pool cost.
    :type minStakePoolCost: Ada
    :param desiredNumberOfStakePools: The desired number of stake pools.
    :type desiredNumberOfStakePools: int
    :param monetaryExpansion: The monetary expansion.
    :type monetaryExpansion: str
    :param treasuryExpansion: The treasury expansion.
    :type treasuryExpansion: str
    :param version: The version.
    :type version: dict
    :param collateralPercentage: The collateral percentage.
    :type collateralPercentage: int
    :param maxCollateralInputs: The maximum collateral inputs.
    :type maxCollateralInputs: int
    :param plutusCostModels: The plutus cost models.
    :type plutusCostModels: dict
    :param scriptExecutionPrices: The script execution prices.
    :type scriptExecutionPrices: dict
    :param maxExecutionUnitsPerTransaction: The maximum execution units per transaction.
    :type maxExecutionUnitsPerTransaction: dict
    :param maxExecutionUnitsPerBlock: The maximum execution units per block.
    :type maxExecutionUnitsPerBlock: dict
    :param maxValueSize: The maximum value size.
    :type maxValueSize: dict
    :param extraEntropy: The extra entropy.
    :type extraEntropy: str
    :param maxTransactionSize: The maximum transaction size.
    :type maxTransactionSize: dict
    :param federatedBlockProductionRatio: The federated block production ratio.
    :type federatedBlockProductionRatio: str
    :param maximumReferenceScriptsSize: The maximum reference scripts size.
    :type maximumReferenceScriptsSize: dict
    :param minFeeReferenceScripts: The minimum fee reference scripts.
    :type minFeeReferenceScripts: dict
    :param stakePoolVotingThresholds: The stake pool voting thresholds.
    :type stakePoolVotingThresholds: dict
    :param delegateRepresentativeVotingThresholds: The delegate representative voting thresholds.
    :type delegateRepresentativeVotingThresholds: dict
    :param constitutionalCommitteeMinSize: The constitutional committee minimum size.
    :type constitutionalCommitteeMinSize: int
    :param constitutionalCommitteeMaxTermLength: The constitutional committee maximum term length.
    :type constitutionalCommitteeMaxTermLength: int
    :param governanceActionLifetime: The governance action lifetime.
    :type governanceActionLifetime: int
    :param governanceActionDeposit: The governance action deposit.
    :type governanceActionDeposit: Ada
    :param delegateRepresentativeDeposit: The delegate representative deposit.
    :type delegateRepresentativeDeposit: Ada
    :param delegateRepresentativeMaxIdleTime: The delegate representative maximum idle time.
    :type delegateRepresentativeMaxIdleTime: int
    """

    def __init__(
        self,
        minFeeCoefficient: int,
        minFeeConstant: Union[int, Ada],
        minUtxoDepositCoefficient: int,
        minUtxoDepositConstant: Union[int, Ada],
        maxBlockBodySize: dict,
        maxBlockHeaderSize: dict,
        stakeCredentialDeposit: Ada,
        stakePoolDeposit: Union[int, Ada],
        stakePoolRetirementEpochBound: int,
        stakePoolPledgeInfluence: str,
        minStakePoolCost: Union[int, Ada],
        desiredNumberOfStakePools: int,
        monetaryExpansion: str,
        treasuryExpansion: str,
        version: dict,
        collateralPercentage: int = None,
        maxCollateralInputs: int = None,
        plutusCostModels: dict = None,
        scriptExecutionPrices: dict = None,
        maxExecutionUnitsPerTransaction: dict = None,
        maxExecutionUnitsPerBlock: dict = None,
        maxValueSize: dict = None,
        extraEntropy: str = None,
        maxTransactionSize: dict = None,
        federatedBlockProductionRatio: str = None,
        maxReferenceScriptsSize: dict = None,
        minFeeReferenceScripts: dict = None,
        stakePoolVotingThresholds: dict = None,
        delegateRepresentativeVotingThresholds: dict = None,
        constitutionalCommitteeMinSize: int = None,
        constitutionalCommitteeMaxTermLength: int = None,
        governanceActionLifetime: int = None,
        governanceActionDeposit: Ada = None,
        delegateRepresentativeDeposit: Ada = None,
        delegateRepresentativeMaxIdleTime: int = None,
    ):
        self.min_fee_coefficient = minFeeCoefficient
        self.min_fee_constant = self._to_ada(minFeeConstant)
        self.min_utxo_deposit_coefficient = minUtxoDepositCoefficient
        self.min_utxo_deposit_constant = self._to_ada(minUtxoDepositConstant)
        self.max_block_body_size = maxBlockBodySize
        self.max_block_header_size = maxBlockHeaderSize
        self.max_transaction_size = maxTransactionSize
        self.max_value_size = maxValueSize
        self.extra_entropy = extraEntropy
        self.stake_credential_deposit = self._to_ada(stakeCredentialDeposit)
        self.stake_pool_deposit = self._to_ada(stakePoolDeposit)
        self.stake_pool_retirement_epoch_bound = stakePoolRetirementEpochBound
        self.stake_pool_pledge_influence = stakePoolPledgeInfluence
        self.min_stake_pool_cost = self._to_ada(minStakePoolCost)
        self.desired_number_of_stake_pools = desiredNumberOfStakePools
        self.federated_block_production_ratio = federatedBlockProductionRatio
        self.monetary_expansion = monetaryExpansion
        self.treasury_expansion = treasuryExpansion
        self.collateral_percentage = collateralPercentage
        self.max_collateral_inputs = maxCollateralInputs
        self.plutus_cost_models = plutusCostModels
        self.script_execution_prices = scriptExecutionPrices
        self.max_execution_units_per_transaction = maxExecutionUnitsPerTransaction
        self.max_execution_units_per_block = maxExecutionUnitsPerBlock
        self.max_ref_script_size = maxReferenceScriptsSize
        self.min_fee_ref_scripts = minFeeReferenceScripts
        self.stake_pool_voting_thresholds = stakePoolVotingThresholds
        self.delegate_representative_voting_thresholds = delegateRepresentativeVotingThresholds
        self.constitutional_committee_min_size = constitutionalCommitteeMinSize
        self.constitutional_committee_max_term_length = constitutionalCommitteeMaxTermLength
        self.governance_action_lifetime = governanceActionLifetime
        self.governance_action_deposit = self._to_ada(governanceActionDeposit)
        self.delegate_representative_deposit = self._to_ada(delegateRepresentativeDeposit)
        self.delegate_representative_max_idle_time = delegateRepresentativeMaxIdleTime
        self.version = version
        self._schema_type = om.ProtocolParameters(
            minFeeCoefficient=self.min_fee_coefficient,
            minFeeConstant=self.min_fee_constant.__dict__(),
            minUtxoDepositCoefficient=self.min_utxo_deposit_coefficient,
            minUtxoDepositConstant=self.min_utxo_deposit_constant.__dict__(),
            maxBlockBodySize=self.max_block_body_size,
            maxBlockHeaderSize=self.max_block_header_size,
            maxTransactionSize=self.max_transaction_size,
            maxValueSize=self.max_value_size,
            extraEntropy=self.extra_entropy,
            stakeCredentialDeposit=self.stake_credential_deposit.__dict__(),
            stakePoolDeposit=self.stake_pool_deposit.__dict__(),
            stakePoolRetirementEpochBound=self.stake_pool_retirement_epoch_bound,
            stakePoolPledgeInfluence=self.stake_pool_pledge_influence,
            minStakePoolCost=self.min_stake_pool_cost.__dict__(),
            desiredNumberOfStakePools=self.desired_number_of_stake_pools,
            federatedBlockProductionRatio=self.federated_block_production_ratio,
            monetaryExpansion=self.monetary_expansion,
            treasuryExpansion=self.treasury_expansion,
            collateralPercentage=self.collateral_percentage,
            maxCollateralInputs=self.max_collateral_inputs,
            plutusCostModels=self.plutus_cost_models,
            scriptExecutionPrices=self.script_execution_prices,
            maxExecutionUnitsPerTransaction=self.max_execution_units_per_transaction,
            maxExecutionUnitsPerBlock=self.max_execution_units_per_block,
            version=self.version,
            minFeeReferenceScripts=self.min_fee_ref_scripts,
            maxReferenceScriptsSize=self.max_ref_script_size,
            stakePoolVotingThresholds=self.stake_pool_voting_thresholds,
            delegateRepresentativeVotingThresholds=self.delegate_representative_voting_thresholds,
            constitutionalCommitteeMinSize=self.constitutional_committee_min_size,
            constitutionalCommitteeMaxTermLength=self.constitutional_committee_max_term_length,
            governanceActionLifetime=self.governance_action_lifetime,
            governanceActionDeposit=(
                self.governance_action_deposit.__dict__()
                if self.governance_action_deposit
                else None
            ),
            delegateRepresentativeDeposit=(
                self.delegate_representative_deposit.__dict__()
                if self.delegate_representative_deposit
                else None
            ),
            delegateRepresentativeMaxIdleTime=self.delegate_representative_max_idle_time,
        )

    def _to_ada(self, value: Union[int, dict, Ada]) -> Ada:
        if value is None:
            return None
        elif isinstance(value, Ada):
            return value
        elif isinstance(value, dict):
            return Ada(value.get("ada").get("lovelace"), is_lovelace=True)
        # elif isinstance(value, int):
        #     return Ada(value)
        else:
            raise InvalidOgmiosParameter(f"Invalid type for value {value}: {type(value)}")


class AlonzoUpdatableParameters:
    """A class representing the Alonzo updatable parameters of the blockchain.

    :param minUtxoDepositCoefficient: The minimum UTXO deposit coefficient.
    :type minUtxoDepositCoefficient: int
    :param collateralPercentage: The collateral percentage.
    :type collateralPercentage: int
    :param plutusCostModels: The plutus cost models.
    :type plutusCostModels: dict
    :param maxCollateralInputs: The maximum collateral inputs.
    :type maxCollateralInputs: int
    :param maxExecutionUnitsPerBlock: The maximum execution units per block.
    :type maxExecutionUnitsPerBlock: dict
    :param maxExecutionUnitsPerTransaction: The maximum execution units per transaction.
    :type maxExecutionUnitsPerTransaction: dict
    :param maxValueSize: The maximum value size.
    :type maxValueSize: dict
    :param scriptExecutionPrices: The script execution prices.
    :type scriptExecutionPrices: dict
    """

    def __init__(
        self,
        minUtxoDepositCoefficient: int,
        collateralPercentage: int,
        plutusCostModels: dict,
        maxCollateralInputs: int,
        maxExecutionUnitsPerBlock: dict,
        maxExecutionUnitsPerTransaction: dict,
        maxValueSize: dict,
        scriptExecutionPrices: dict,
    ):
        self.min_utxo_deposit_coefficient = minUtxoDepositCoefficient
        self.collateral_percentage = collateralPercentage
        self.plutus_cost_models = plutusCostModels
        self.max_collateral_inputs = maxCollateralInputs
        self.max_execution_units_per_block = maxExecutionUnitsPerBlock
        self.max_execution_units_per_transaction = maxExecutionUnitsPerTransaction
        self.max_value_size = maxValueSize
        self.script_execution_prices = scriptExecutionPrices
        self._schema_type = om.UpdatableParameters(
            minUtxoDepositCoefficient=self.min_utxo_deposit_coefficient,
            collateralPercentage=self.collateral_percentage,
            plutusCostModels=self.plutus_cost_models,
            maxCollateralInputs=self.max_collateral_inputs,
            maxExecutionUnitsPerBlock=self.max_execution_units_per_block,
            maxExecutionUnitsPerTransaction=self.max_execution_units_per_transaction,
            maxValueSize=self.max_value_size,
            scriptExecutionPrices=self.script_execution_prices,
        )


class ConwayUpdatableParameters:
    """A class representing the Conway updatable parameters of the blockchain.

    :param stakePoolVotingThresholds: The stake pool voting thresholds.
    :type stakePoolVotingThresholds: dict
    :param delegateRepresentativeVotingThresholds: The delegate representative voting thresholds.
    :type delegateRepresentativeVotingThresholds: dict
    :param constitutionalCommitteeMinSize: The constitutional committee minimum size.
    :type constitutionalCommitteeMinSize: int
    :param constitutionalCommitteeMaxTermLength: The constitutional committee maximum term length.
    :type constitutionalCommitteeMaxTermLength: int
    :param governanceActionLifetime: The governance action lifetime.
    :type governanceActionLifetime: int
    :param governanceActionDeposit: The governance action deposit.
    :type governanceActionDeposit: Ada
    :param delegateRepresentativeDeposit: The delegate representative deposit.
    :type delegateRepresentativeDeposit: Ada
    :param delegateRepresentativeMaxIdleTime: The delegate representative maximum idle time.
    :type delegateRepresentativeMaxIdleTime: int
    """

    def __init__(
        self,
        stakePoolVotingThresholds: dict,
        delegateRepresentativeVotingThresholds: dict,
        constitutionalCommitteeMinSize: int,
        constitutionalCommitteeMaxTermLength: int,
        governanceActionLifetime: int,
        governanceActionDeposit: Ada,
        delegateRepresentativeDeposit: Ada,
        delegateRepresentativeMaxIdleTime: int,
    ):
        self.stake_pool_voting_thresholds = stakePoolVotingThresholds
        self.delegate_representative_voting_thresholds = delegateRepresentativeVotingThresholds
        self.constitutional_committee_min_size = constitutionalCommitteeMinSize
        self.constitutional_committee_max_term_length = constitutionalCommitteeMaxTermLength
        self.governance_action_lifetime = governanceActionLifetime
        self.governance_action_deposit = governanceActionDeposit
        self.delegate_representative_deposit = delegateRepresentativeDeposit
        self.delegate_representative_max_idle_time = delegateRepresentativeMaxIdleTime
        self._schema_type = om.UpdatableParameters1(
            stakePoolVotingThresholds=self.stake_pool_voting_thresholds,
            delegateRepresentativeVotingThresholds=self.delegate_representative_voting_thresholds,
            constitutionalCommitteeMinSize=self.constitutional_committee_min_size,
            constitutionalCommitteeMaxTermLength=self.constitutional_committee_max_term_length,
            governanceActionLifetime=self.governance_action_lifetime,
            governanceActionDeposit=self.governance_action_deposit,
            delegateRepresentativeDeposit=self.delegate_representative_deposit,
            delegateRepresentativeMaxIdleTime=self.delegate_representative_max_idle_time,
        )


class GovernanceProtocolParameters:
    """A class representing the governance protocol parameters of the blockchain.

    NOTE: This data structure is currently missing from the ogmios schema, so strict
    type checking is not enforced. This will be updated when the schema is updated.

    :param stakePoolVotingThresholds: The stake pool voting thresholds.
    :type stakePoolVotingThresholds: dict
    :param delegateRepresentativeVotingThresholds: The delegate representative voting thresholds.
    :type delegateRepresentativeVotingThresholds: dict
    :param constitutionalCommitteeMinSize: The constitutional committee minimum size.
    :type constitutionalCommitteeMinSize: int
    :param constitutionalCommitteeMaxTermLength: The constitutional committee maximum term length.
    :type constitutionalCommitteeMaxTermLength: int
    :param governanceActionLifetime: The governance action lifetime.
    :type governanceActionLifetime: int
    :param governanceActionDeposit: The governance action deposit.
    :type governanceActionDeposit: Ada
    :param delegateRepresentativeDeposit: The delegate representative deposit.
    :type delegateRepresentativeDeposit: Ada
    :param delegateRepresentativeMaxIdleTime: The delegate representative maximum idle time.
    :type delegateRepresentativeMaxIdleTime: int
    """

    def __init__(
        self,
        stakePoolVotingThresholds: dict,
        delegateRepresentativeVotingThresholds: dict,
        constitutionalCommitteeMinSize: int,
        constitutionalCommitteeMaxTermLength: int,
        governanceActionLifetime: int,
        governanceActionDeposit: Ada,
        delegateRepresentativeDeposit: Ada,
        delegateRepresentativeMaxIdleTime: int,
    ):
        self.stake_pool_voting_thresholds = stakePoolVotingThresholds
        self.delegate_representative_voting_thresholds = delegateRepresentativeVotingThresholds
        self.constitutional_committee_min_size = constitutionalCommitteeMinSize
        self.constitutional_committee_max_term_length = constitutionalCommitteeMaxTermLength
        self.governance_action_lifetime = governanceActionLifetime
        self.governance_action_deposit = governanceActionDeposit
        self.delegate_representative_deposit = delegateRepresentativeDeposit
        self.delegate_representative_max_idle_time = delegateRepresentativeMaxIdleTime
        self._schema_type = None



import pytest

from ogmios.errors import ResponseError
from ogmios.datatypes import Origin, Point

from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_AcquireLedgerState_tip(client):
    # Get tip
    _, tip, _ = client.find_intersection.execute([Origin()])

    id_str = "My ID string"
    point, id = client.acquire_ledger_state.execute(tip.to_point(), id_str)

    assert isinstance(point, Point)
    assert tip.to_point() == point
    assert id == id_str


def test_AcquireLedgerState_origin(client):
    # Get tip
    point, _, _ = client.find_intersection.execute([Origin()])

    # Origin is too old to acquire ledger state
    with pytest.raises(ResponseError):
        _, _ = client.acquire_ledger_state.execute(point)



import ogmios

"""
This example shows how to use the ogmios-python library to asynchronously execute
multiple queries. This is useful for applications that need to query the chain
frequently.
"""

batch_size = 100


def async_queries():
    with ogmios.Client() as client:
        for i in range(batch_size):
            client.query_block_height.send(id=i)

        results = []
        for i in range(batch_size):
            results.append(client.query_block_height.receive())

        results.sort(key=lambda tup: tup[1])
        for height, id in results:
            print(f"Block height: {height} (ID: {id})")


if __name__ == "__main__":
    async_queries()



import json
import base64
from cardano_tools import WalletHTTP
import ogmios


"""
This example shows how to build and sign a transaction using the CardanoTools wallet API. This requires
cardano-wallet to be running.

https://gitlab.com/viperscience/cardano-tools
https://github.com/cardano-foundation/cardano-wallet
"""

WALLET_URL = "http://localhost"
WALLET_PORT = 8090
WALLET_NAME = "my-wallet"
MNEMONIC_FILE = "./tests/test_wallet/test.seed"

# Unsafe to use in production - read from environment variable instead
WALLET_PASSPHRASE = "my-passphrase"


cw_api = WalletHTTP(WALLET_URL, WALLET_PORT)

# Import wallet if it hasn't been already
if not (wallet_id := cw_api.get_wallet_by_name(WALLET_NAME).get("id")):
    with open(MNEMONIC_FILE, "r") as f:
        mnemonic = f.read()
        cw_api.create_wallet(WALLET_NAME, mnemonic.strip().split(" "), WALLET_PASSPHRASE)
        wallet_id = cw_api.get_wallet_by_name(WALLET_NAME).get("id")

# Send 2 ADA and a native asset (CHOC) in quantity of 2000 to an address.
target_addr = "addr_test1wztjpvnxahymaejx9qa9c3xddkt3ujhjykmt83uu8xfvprq7hxv53"
lovelace_amt = 2 * 1e6
token_policy = "57fca08abbaddee36da742a839f7d83a7e1d2419f1507fcbf3916522"
asset_name = "CHOC"
token_qty = 2000

payload = json.loads(
    f"""{{
        "payments": [
            {{
                "address": "{target_addr}",
                "amount": {{
                    "quantity": {lovelace_amt},
                    "unit": "lovelace"
                }},
                "assets": [
                    {{
                        "policy_id": "{token_policy}",
                        "asset_name": "{asset_name}",
                        "quantity": {token_qty}
                    }}
                ]
            }}
        ],
        "withdrawal": "self",
        "validity_interval": {{
            "invalid_hereafter": {{
                "quantity": 3600,
                "unit": "second"
            }}
        }},
        "encoding": "base16"
    }}"""
)

tx = cw_api.construct_transaction(wallet_id, payload)
encoded_tx = tx.get("transaction")
signed_tx = cw_api.sign_transaction(wallet_id, WALLET_PASSPHRASE, encoded_tx)
tx_cbor = base64.b64decode(signed_tx.get("transaction")).hex()

with ogmios.Client() as client:
    tx_id, _ = client.submit_transaction.execute(tx_cbor)



from typing import Union, Dict

from ogmios.errors import InvalidResponseError
import ogmios.model.cardano_model as cm
import ogmios.model.ogmios_model as om
from ogmios.datatypes import Origin, Point, Tip, Block
import ogmios.model.model_map as mm

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def parse_PointOrOrigin(resp: Union[Dict, str]) -> Union[Point, Origin]:
    """Parse a response that contains either a point or an origin.

    :param resp: The response to parse.
    :type resp: str | dict
    :raises InvalidResponseError: If the response is not a valid point or origin.
    :return: The parsed point or origin.
    """
    if isinstance(resp, str):
        assert resp == om.Origin.origin.value
        return Origin()
    elif isinstance(resp, dict) and set(resp.keys()) == {"slot", "id"}:
        return Point(slot=resp.get("slot"), id=resp.get("id"))
    else:
        raise InvalidResponseError(f"Invalid point: {resp}")


def parse_TipOrOrigin(resp: Union[Dict, str]) -> Union[Tip, Origin]:
    """Parse a response that contains either a tip or an origin.

    :param resp: The response to parse.
    :type resp: str | dict
    :raises InvalidResponseError: If the response is not a valid tip or origin.
    :return: The parsed tip or origin.
    """
    if isinstance(resp, str):
        assert resp == om.Origin.origin.value
        return Origin()
    elif isinstance(resp, dict) and set(resp.keys()) == {"slot", "height", "id"}:
        return Tip(slot=resp.get("slot"), height=resp.get("height"), id=resp.get("id"))
    else:
        raise InvalidResponseError(f"Invalid tip: {resp}")


def parse_BlockHeightOrOrigin(resp: Union[int, str]) -> Union[int, Origin]:
    """Parse a response that contains either a point or an origin.

    :param resp: The response to parse.
    :type resp: str | dict
    :raises InvalidResponseError: If the response is not a valid point or origin.
    :return: The parsed point or origin.
    """
    if isinstance(resp, str):
        assert resp == om.Origin.origin.value
        return Origin()
    elif isinstance(resp, int):
        return resp
    else:
        raise InvalidResponseError(f"Invalid block height response: {resp}")


def parse_Block(resp: dict) -> Union[cm.BlockEBB, cm.BlockBFT, cm.BlockPraos]:
    """Parse a response that contains a block.

    :param resp: The response to parse.
    :type resp: dict
    :raises InvalidResponseError: If the response is not a valid block.
    :return: The parsed block.
    """
    if btype := resp.get("type"):
        if (
            btype == mm.Types.ebb.value
            and (era := resp.get("era")) is not None
            and (id := resp.get("id")) is not None
            and (ancestor := resp.get("ancestor")) is not None
            and (height := resp.get("height")) is not None
        ):
            return Block(blocktype=btype, era=era, id=id, ancestor=ancestor, height=height)
        elif (
            btype == mm.Types.bft.value
            and (era := resp.get("era"))
            and (id := resp.get("id")) is not None
            and (ancestor := resp.get("ancestor")) is not None
            and (height := resp.get("height")) is not None
            and (slot := resp.get("slot")) is not None
            and (size := resp.get("size")) is not None
            and (protocol := resp.get("protocol")) is not None
            and (issuer := resp.get("issuer")) is not None
            and (delegate := resp.get("delegate")) is not None
        ):
            # Optional parameters
            transactions = resp.get("transactions")
            opcert = resp.get("operationalCertificate")

            return Block(
                blocktype=btype,
                era=era,
                id=id,
                ancestor=ancestor,
                height=height,
                slot=slot,
                size=size,
                transactions=transactions,
                operationalCertificates=opcert,
                protocol=protocol,
                issuer=issuer,
                delegate=delegate,
            )
        elif (
            btype == mm.Types.praos.value
            and (era := resp.get("era")) is not None
            and (id := resp.get("id")) is not None
            and (ancestor := resp.get("ancestor")) is not None
            and (height := resp.get("height")) is not None
            and (slot := resp.get("slot")) is not None
            and (size := resp.get("size")) is not None
            and (protocol := resp.get("protocol")) is not None
            and (issuer := resp.get("issuer")) is not None
        ):
            nonce = resp.get("nonce")
            transactions = resp.get("transactions")

            return Block(
                blocktype=btype,
                era=era,
                id=id,
                ancestor=ancestor,
                nonce=nonce,
                height=height,
                size=size,
                slot=slot,
                transactions=transactions,
                protocol=protocol,
                issuer=issuer,
            )
    raise InvalidResponseError(f"Invalid block information: {resp}")



import ogmios
import pycardano as pyc

"""
This example shows how to build and sign a transaction using PyCardano.

https://github.com/Python-Cardano/pycardano
"""

network = pyc.Network.TESTNET

# Read keys to memory
psk = pyc.PaymentExtendedSigningKey.load("./tests/test_wallet/test_addr0.skey")
ssk = pyc.StakeExtendedSigningKey.load("./tests/test_wallet/test_stake.skey")

pvk = pyc.PaymentVerificationKey.from_signing_key(psk)
svk = pyc.StakeVerificationKey.from_signing_key(ssk)

# Derive an address from payment verification key and stake verification key
address = pyc.Address(pvk.hash(), svk.hash(), network)

# Create an Ogmios chain context. In this example, we will use preprod network.
context = pyc.OgmiosV6ChainContext("localhost", 1337)

# Create a transaction builder
builder = pyc.TransactionBuilder(context)

# Tell the builder that transaction input will come from a specific address, assuming that there are some ADA and native
# assets sitting at this address. "add_input_address" could be called multiple times with different address.
builder.add_input_address(address)

# Get all UTxOs currently sitting at this address
utxos = context.utxos(address)

# We can also tell the builder to include a specific UTxO in the transaction.
# Similarly, "add_input" could be called multiple times.
builder.add_input(utxos[0])

builder.add_output(
    pyc.TransactionOutput(
        pyc.Address.from_primitive(
            "addr_test1vrm9x2zsux7va6w892g38tvchnzahvcd9tykqf3ygnmwtaqyfg52x"
        ),
        pyc.Value.from_primitive(
            [
                1500000,
            ]
        ),
    )
)

# We can add multiple outputs, similar to what we can do with inputs.
# Send 2 ADA and a native asset (CHOC) in quantity of 200 to ourselves
builder.add_output(
    pyc.TransactionOutput(
        address,
        pyc.Value.from_primitive(
            [
                2000000,
                {
                    bytes.fromhex(
                        "57fca08abbaddee36da742a839f7d83a7e1d2419f1507fcbf3916522"  # Policy ID
                    ): {
                        b"CHOC": 200  # Asset name and amount
                    }
                },
            ]
        ),
    )
)

# Create final signed transaction
signed_tx = builder.build_and_sign([psk], change_address=address)
print(signed_tx)

# Submit signed transaction to the network
context.submit_tx(signed_tx)



from datetime import datetime

from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_QueryRewardAccountSummaries_key_empty(client):
    # This test relies on this address having a balance on the testnet
    stake_key = "stake_test1urlvxquq2pm8teqed7ek4sgjuc0zrx07z9czmy5fvwey92cq8c5dh"
    id_str = "My ID string"
    result, id = client.query_reward_account_summaries.execute(keys=[stake_key], id=id_str)

    assert result == {}
    assert id == id_str


def test_QueryRewardAccountSummaries_key(client):
    # This test relies on this address having a balance on the testnet
    stake_key = "stake_test1uq9947tw3wg5y7l0wmc59eyuz2fc4p76m6qjnwmw5ss58vsmzhs0n"
    id_str = "My ID string"
    result, id = client.query_reward_account_summaries.execute(keys=[stake_key], id=id_str)

    expected_keys = ["delegate", "rewards"]
    assert isinstance(result, dict)
    for key in expected_keys:
        assert key in result.get(list(result.keys())[0]).keys()
    assert id == id_str


def test_QueryRewardAccountSummaries_script_empty(client):
    # This test relies on this address having a balance on the testnet
    script_key = "stake_test1urlvxquq2pm8teqed7ek4sgjuc0zrx07z9czmy5fvwey92cq8c5dh"
    id_str = "My ID string"
    result, id = client.query_reward_account_summaries.execute(scripts=[script_key], id=id_str)

    assert result == {}
    assert id == id_str



import json
from typing import no_type_check, Type, Any
from pydantic.v1 import BaseModel, ValidationError, StrBytes, Protocol

from ogmios.logger import logger


class BaseModelNoException(BaseModel):
    def __init__(__pydantic_self__, **data: Any) -> None:
        try:
            super(BaseModelNoException, __pydantic_self__).__init__(**data)
        except ValidationError as pve:
            logger.warning(
                f'pydantic warning:  __init__ failed to validate:\n {json.dumps(data, indent=4)}\n'
            )
            logger.warning(f'This is the original exception:\n{pve.json()}')

    @no_type_check
    def __setattr__(self, name, value):
        try:
            return super(BaseModelNoException, self).__setattr__
        except ValidationError as pve:
            logger.warning(
                f'pydantic warning:  __setattr__ failed to validate:\n {json.dumps({name: value}, indent=4)}'
            )
            logger.warning(f'This is the original exception:\n{pve.json()}')
            return None

    @classmethod
    def parse_obj(cls: Type['Model'], obj: Any) -> 'Model':
        try:
            return super(BaseModelNoException, cls).parse_obj(obj)
        except ValidationError as pve:
            logger.warning(
                f'pydantic warning:  parse_obj failed to validate:\n {json.dumps(obj, indent=4)}'
            )
            logger.warning(f'This is the original exception:\n{pve.json()}')
            return None

    @classmethod
    def parse_raw(
        cls: Type['Model'],
        b: StrBytes,
        *,
        content_type: str = None,
        encoding: str = 'utf8',
        proto: Protocol = None,
        allow_pickle: bool = False,
    ) -> 'Model':
        try:
            return super(BaseModelNoException, cls).parse_raw(
                b=b,
                content_type=content_type,
                encoding=encoding,
                proto=proto,
                allow_pickle=allow_pickle,
            )
        except ValidationError as pve:
            logger.warning(f'pydantic warning:  parse_raw failed to validate:\n {b}')
            logger.warning(f'This is the original exception:\n{pve.json()}')
            return None



"""This file contains convenient mappings from ogmios.datatypes to ogmios.model.ogmios_model.
    These mappings combine logical types from the ogmios model that are separated in the schema into
    combined enums that are easier to work with.
"""

from enum import Enum

import ogmios.model.ogmios_model as om


class Types(Enum):
    ebb = om.Type.ebb.value
    bft = om.Type1.bft.value
    praos = om.Type2.praos.value
    stakeDelegation = om.Type3.stakeDelegation.value
    stakeCredentialRegistration = om.Type4.stakeCredentialRegistration.value
    stakeCredentialDeregistration = om.Type5.stakeCredentialDeregistration.value
    stakePoolRegistration = om.Type6.stakePoolRegistration.value
    stakePoolRetirement = om.Type7.stakePoolRetirement.value
    genesisDelegation = om.Type8.genesisDelegation.value
    constitutionalCommitteeDelegation = om.Type9.constitutionalCommitteeDelegation.value
    constitutionalCommitteeRetirement = om.Type10.constitutionalCommitteeRetirement.value
    delegateRepresentativeRegistration = om.Type11.delegateRepresentativeRegistration.value
    delegateRepresentativeUpdate = om.Type12.delegateRepresentativeUpdate.value
    delegateRepresentativeRetirement = om.Type13.delegateRepresentativeRetirement.value
    registered = om.Type14.registered.value
    noConfidence = om.Type15.noConfidence.value
    abstain = om.Type16.abstain.value
    protocolParametersUpdate = om.Type17.protocolParametersUpdate.value
    hardForkInitiation = om.Type18.hardForkInitiation.value
    treasuryTransfer = om.Type19.treasuryTransfer.value
    treasuryWithdrawals = om.Type20.treasuryWithdrawals.value
    constitutionalCommittee = om.Type21.constitutionalCommittee.value
    constitution = om.Type22.constitution.value
    information = om.Type24.information.value
    ipAddress = om.Type25.ipAddress.value
    hostname = om.Type26.hostname.value


class Method(Enum):
    findIntersection = om.Method.findIntersection.value
    nextBlock = om.Method4.nextBlock.value
    submitTransaction = om.Method6.submitTransaction.value
    evaluateTransaction = om.Method10.evaluateTransaction.value
    acquireLedgerState = om.Method14.acquireLedgerState.value
    releaseLedgerState = om.Method17.releaseLedgerState.value
    queryLedgerState_epoch = om.Method19.queryLedgerState_epoch.value
    queryLedgerState_eraStart = om.Method19.queryLedgerState_eraStart.value
    queryLedgerState_eraSummaries = om.Method19.queryLedgerState_eraSummaries.value
    queryLedgerState_liveStakeDistribution = (
        om.Method19.queryLedgerState_liveStakeDistribution.value
    )
    queryLedgerState_projectedRewards = om.Method19.queryLedgerState_projectedRewards.value
    queryLedgerState_protocolParameters = om.Method19.queryLedgerState_protocolParameters.value
    queryLedgerState_proposedProtocolParameters = (
        om.Method19.queryLedgerState_proposedProtocolParameters
    ).value
    queryLedgerState_rewardAccountSummaries = (
        om.Method19.queryLedgerState_rewardAccountSummaries.value
    )
    queryLedgerState_rewardsProvenance = om.Method19.queryLedgerState_rewardsProvenance.value
    queryLedgerState_stakePools = om.Method19.queryLedgerState_stakePools.value
    queryLedgerState_utxo = om.Method19.queryLedgerState_utxo.value
    queryLedgerState_tip = om.Method19.queryLedgerState_tip.value
    queryLedgerState_constitution = om.Method19.queryLedgerState_constitution.value
    queryLedgerState_constitutionalCommittee = (
        om.Method19.queryLedgerState_constitutionalCommittee.value
    )
    queryLedgerState_treasuryAndReserves = om.Method19.queryLedgerState_treasuryAndReserves.value
    queryNetwork_blockHeight = om.Method52.queryNetwork_blockHeight.value
    queryNetwork_genesisConfiguration = om.Method54.queryNetwork_genesisConfiguration.value
    queryNetwork_startTime = om.Method57.queryNetwork_startTime.value
    queryNetwork_tip = om.Method59.queryNetwork_tip.value
    acquireMempool = om.Method61.acquireMempool.value
    nextTransaction = om.Method63.nextTransaction.value
    hasTransaction = om.Method64.hasTransaction.value
    sizeOfMempool = om.Method64.sizeOfMempool.value
    releaseMempool = om.Method64.releaseMempool.value



import pytest

from ogmios.chainsync.FindIntersection import FindIntersection
from ogmios.errors import InvalidMethodError, InvalidResponseError, ResponseError
from ogmios.datatypes import Origin, Point, Tip

from tests.ogmios.test_fixtures import client, last_byron_block_point

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_FindIntersection_origin(client):
    id_str = "My ID string"
    intersection, tip, id = client.find_intersection.execute([Origin()], id_str)

    assert isinstance(intersection, Origin)
    assert isinstance(tip, Tip)
    assert id == id_str


def test_FindIntersection_last_byron_block(client, last_byron_block_point):
    """This test assumes we are connected to the preprod network"""
    intersection, _, _ = client.find_intersection.execute([last_byron_block_point, Origin()])
    assert intersection == last_byron_block_point


def test_FindIntersection_tip(client):
    """This test assumes we are connected to the preprod network"""
    # Find origin just to get the tip
    _, tip, _ = client.find_intersection.execute([Origin()])
    point = Point(slot=tip.slot, id=tip.id)

    # Now find intersection with the tip
    intersection, _, _ = client.find_intersection.execute([point, Origin()])
    assert intersection == point


def test_FindIntersection_invalid_block(client):
    point = Point(slot=1, id="1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef")
    with pytest.raises(ResponseError):
        _, _, _ = client.find_intersection.execute([point])


def test_FindIntersection_parser_method_error():
    resp_bad_method = {
        "jsonrpc": "2.0",
        "method": "badMethodName",
        "result": {
            "intersection": {
                "slot": 18446744073709552000,
                "id": "c248757d390181c517a5beadc9c3fe64bf821d3e889a963fc717003ec248757d",
            },
            "tip": {
                "slot": 18446744073709552000,
                "id": "c248757d390181c517a5beadc9c3fe64bf821d3e889a963fc717003ec248757d",
                "height": 18446744073709552000,
            },
        },
        "id": "1234",
    }
    with pytest.raises(InvalidMethodError):
        FindIntersection._parse_FindIntersection_response(resp_bad_method)


def test_FindIntersection_response_error():
    resp_bad_key = {
        "jsonrpc": "2.0",
        "method": "findIntersection",
        "result": {
            "badKeyName": {
                "slot": 18446744073709552000,
                "id": "c248757d390181c517a5beadc9c3fe64bf821d3e889a963fc717003ec248757d",
            },
            "tip": {
                "slot": 18446744073709552000,
                "id": "c248757d390181c517a5beadc9c3fe64bf821d3e889a963fc717003ec248757d",
                "height": 18446744073709552000,
            },
        },
        "id": "1234",
    }
    with pytest.raises(InvalidResponseError):
        FindIntersection._parse_FindIntersection_response(resp_bad_key)
